#ifndef _graph_codegen_h_
#define _graph_codegen_h_

// (C) by Thalion Graphic Solutions e.K. 2001
//
// Permissions to copy, modify and redistribute
// granted, according to the conditions defined
// in the Gnu Public License (GPL).
// Please read the file gpl.txt provided with this
// distribution.

#include <graph/runtime.h>

#include <vector>

namespace graph {

   //
   // This class provides convenience methods for compilers
   // to create code for the graph stack machine.
   //
   class CodeGen {

      public:

         //
         // Errors.
         //
         // Thrown, if the number of constant entries exceeds
         // the runtime environment's data buffer size.
         //
         class datasize_exceeded {};
      
         //
	 // Creates a code generator instance for a runtime instance.
	 // The code generator will add it's code to the runtime's data
	 // segment, and allows the user to retrieve function instances
	 // after each completed definition.
	 //
         CodeGen(Runtime&);
      
         //
	 // Start a new function definition.
	 // Returns the data segment pointer to this function.
	 // Synopsis: unsigned fp = <codegen>.beginFunction(<function_name>,<function_signature>);
	 //
	 unsigned beginFunction(const char *name,const char *signature = "(f)f",unsigned argc = 1);
         
	 //
	 // Add an instruction to a function.
	 // Returns the current address.
	 // Synopsis: unsigned addr = <codegen>.addIns(<instruction>);
	 //
	 unsigned addIns(Runtime::ins_t);
	 
	 //
	 // Add an instruction plus parameter to a function.
	 // Returns the current address.
	 // Synopsis: unsigned addr = <codegen>.addIns(<instruction>,<parameter>);
	 //
	 unsigned addIns(Runtime::ins_t,unsigned);

         //
         // Sets the label of branch instruction at address.
         // Synopsis: <codegen>->setLabel(<ins-address>,<branchlabel>);
         void setLabel(unsigned,unsigned);

         //
	 // End a function definition.
         // Throws datasize_exceeded, if the data buffer entry fails
	 // (create a Runtime instance with larger data space, if so).
	 // Synopsis: <codegen>->endFunction();  
         void endFunction();
	 
	 //
	 // Return the current function (NULL, if invoked before the
	 // first endFunction().
	 // Synopsis: Function f = <codegen>.getFunction();
	 //
         const Function *getFunction();

         //
         // Returns the current function's name.
         //
         const char *getName() const;
	 
	 //
	 // Adds a constant.
         // Throws datasize_exceeded, if the data buffer entry fails
	 // (create a Runtime instance with larger data space, if so).
	 //
	 unsigned constant(double);
	 unsigned constant(long);
	 
         //
         // Resets the code generator, ie. sets back the 
         // ds pointer to 0. All data in the runtime environment
         // will be overwritten by further calls to the code generation
         // api.
         //
         void reset();

      private:
      
         // The runtime environment.
         Runtime&                    env_;
	 
	 // Data segment pointer.
	 unsigned                    dp_;

         // Current function state: function, code segment container, name, sig, argument count
	 // and data segment pointer.
	 const Function             *cf_;
         std::vector<Runtime::Ins>   cf_cs_;	 
	 char                       *cf_name_;
	 char                       *cf_signature_;
	 unsigned                    cf_argc_;
	 unsigned                    cf_dp_;
	 
   };

}

inline unsigned graph::CodeGen::constant(double d)
{
   if( env_.setData(dp_,d) ) {
      return dp_++;
   }
   else {
      throw datasize_exceeded();
   }
}

inline unsigned graph::CodeGen::constant(long l)
{
   if( env_.setData(dp_,l) ) {
      return dp_++;
   }
   else {
      throw datasize_exceeded();
   }
}

inline const graph::Function *graph::CodeGen::getFunction()
{
   return cf_;
}

inline const char *graph::CodeGen::getName() const
{
   return cf_name_;
}

inline void graph::CodeGen::setLabel(unsigned addr,unsigned label)
{
   cf_cs_[addr].par_ = label;
}

#endif
