#ifndef _graph_graph3d_h_
#define _graph_graph3d_h_

// (C) by Thalion Graphic Solutions e.K. 2001
//
// Permissions to copy, modify and redistribute
// granted, according to the conditions defined
// in the Gnu Public License (GPL).
// Please read the file gpl.txt provided with this
// distribution.

#include <GL/gl.h>

#include <graph/refcounted.h>

namespace graph {

   template <int GW = 100,int GH = 100>
   class Graph3d;

   //
   // A function taking 2 arguments.
   //
   class Function2: public RefCounted {
      public:
         virtual double operator () (double,double) = 0;
   };

   //
   // A function taking 3 arguments.
   //
   class Function3: public RefCounted {
      public:
         virtual double operator () (double,double,double) = 0;
   };

   //
   // Base class for all things that can be displayed
   // by the Graph3d class.
   //
   template <int GW = 100,int GH = 100>
   class Graphable {
      public:
         virtual Graph3d<GW,GH> *getGraph(double x0=-1,
                                          double x1=1,
                                          double y0=-1,
                                          double y1=1,
                                          int mode=1) = 0;
   };

   //
   // A graphable function taking 2 arguments.
   //
   template <int GW = 100,int GH = 100>
   class GraphableFunction2:
      public Function2,
      public Graphable<GW,GH> {
      public:
         virtual double operator () (double,double) = 0;
         virtual Graph3d<GW,GH> *getGraph(double x0=-1,
                                          double x1=1,
                                          double y0=-1,
                                          double y1=1,
                                          int mode=1);
   };

   //
   // A graphable function taking 3 arguments.
   // The clone operation is required here as well.
   //
   template <int GW = 100,int GH = 100>
   class GraphableFunction3:
      public Function3,
      public Graphable<GW,GH> {
      public:
         virtual double operator () (double,double,double) = 0;
         virtual Graph3d<GW,GH> *getGraph(double x0=-1,
                                          double x1=1,
                                          double y0=-1,
                                          double y1=1,
                                          int mode=1);
	 virtual ~GraphableFunction3();
   };

   //
   // Wraps a 3-argument function.
   //
   template <int GW = 100,int GH = 100>
   class Function3Wrapper: public GraphableFunction2<GW,GH> {
      public:
         Function3Wrapper(GraphableFunction3<GW,GH>&);
         double operator () (double,double);
         void setZ(double);
      protected:	 
         virtual ~Function3Wrapper();
      private:
         GraphableFunction3<GW,GH> *f_;
         double                     z_;
   };
   
   //
   // An instance of this class is used as default
   // color function for r,g and b.
   //
   class DefaultRGBFunc: public Function3 {
      public:
         double operator () (double,double,double);
         static DefaultRGBFunc *getInstance();
      private:
         static DefaultRGBFunc *instance_;
   };

   //
   // The class encapsulates visualization of a function with 2 or
   // 3 parameters.
   //
   template <int GW,int GH>
   class Graph3d {

      public:
         //
	 // Supported modes.
         //
	 enum mode_bits {
	    POINTS = 0x01,
	    LINES  = 0x02,
	    SMOOTH = 0x04,
	    LIGHTS = 0x08,
	    AXES   = 0x10,
            DLISTS = 0x20
	 };
	 
	 //
	 // Errors.
	 //
	 class illegal_argument_error {};
      
         //
	 // Initialization.
	 //
	 
         // Create the graph.
         Graph3d(GraphableFunction2<GW,GH>& f,
                 double x0=-1,
                 double x1=1,
                 double y0=-1,
                 double y1=1,
                 int mode=POINTS|DLISTS);
         Graph3d(GraphableFunction3<GW,GH>& f,
                 double x0=-1,
                 double x1=1,
                 double y0=-1,
                 double y1=1,
                 int mode=POINTS|DLISTS);

         // Initialize it.
         void init();

         // Optionally initialize other than
         // default color settings.
         void initColors(float l0r,float l0g,float l0b,
                         float l1r,float l1g,float l1b,
                         float bgr,float bgg,float bgb);
	 
	 //
	 // Painting.
	 //
	 
         // Display it.
         void display();
         // React on window resizes.
         void reshape(int,int);
	 
	 //
	 // Transforms.
	 //
	 
         // Rotate around x,y,z axis.
         void rotateX(double);
         void rotateY(double);
         void rotateZ(double);

	 // Rotate around axis determined by screen coordinates
	 void screenRotate(int,int,int,int);

	 // Zoom in or out.
	 void zoom(double);
	 
	 //
	 // Getters.
	 //
	 
	 // Light #n rgb values.
	 float getLightRedValue(int) const;
	 float getLightGreenValue(int) const;
	 float getLightBlueValue(int) const;
	 
	 // Plot ranges.
	 double getRangeX0() const;
	 double getRangeX1() const;
	 double getRangeY0() const;
	 double getRangeY1() const;
	 
	 // Is it a f(x,y,z)-style function?
	 bool hasZ() const;

         // Test for mode flags.
         int mode(int) const;
	 
	 //
	 // Setters
	 //

         // Set plot X and Y ranges.
         void setRangeX(double,double);
         void setRangeY(double,double);

         // Set Z value for a f(x,y,z) style function.
         // Will be ignored for binary function.
         void setZ(double);

         // Set background/clear color.
         void setClearColor(float,float,float);

         // Set light #n rgb color.
         void setLightColor(int,float,float,float);

         // Sets mode flags.
	 void setMode(int);
	 
	 // Unsets all mode flags.
	 void setDefaultMode();

         // Set functions for vertex R,G,B computation.
         void setRFunction(Function3&);
         void setGFunction(Function3&);
         void setBFunction(Function3&);
	 
	 // Unsets mode flags.
	 void unsetMode(int);

         // Destroy the world.
         virtual ~Graph3d();

      protected:
         //
         // Methods.
         //

         // Initialize lighting parameters.
         void initLights();
         // Initialize the graph.
         void initGraphValues();
         void initGraphIndices();

         // Initialize the GL scene settings.
         void initSceneSettings();

         // Create display list.
         void createDisplayList();

         // Draw the components immediately.
         void drawAxes();
         void drawGrid(); 

      private:
         //
         // Methods.
         //
	 
         // Set lighting on or off.
         void setLighting(bool flag);

         // Set points mode on.
         void setPoints();
         // Set wireframe mode on.
         void setLines();
         // Set smooth shading mode on.
         void setSmooth();
	 
         //
         // Attributes.
         //

         // The function.
         GraphableFunction2<GW,GH> *function_;

         // GL data arrays.
         double                     vertices_[GW * GH * 3];
         double                     normals_[(GW - 1) * (GH - 1) * 3];
         double                     colors_[GW * GH * 3];
         GLuint                     indices_[(GW - 1) * (GH - 1) * 4];

         // Vertex color functions. 
         Function3                 *rfunc_;
         Function3                 *gfunc_;
         Function3                 *bfunc_;

         // Range.
         double                     x0_, x1_, y0_, y1_;

         // Computed z range: min and max of function.
         double                     min_, max_;

         // RGBA light color arrays.
         float                     *light0_ambient_;
         float                     *light0_color_;
         float                     *light1_ambient_;
         float                     *light1_color_;

         // BG color r,g,b values.
         float                      bg_r_,bg_g_,bg_b_;

         // Geometry.
         //
         // Current rotation around x,y,z axis.
         double                     rx_, ry_, rz_;
	 // Translation along z (zoom).
	 double                     tz_;
         // Modelview matrix.
         float                      modelview_[16];

         // Display list handles.
         GLuint                     hgraph_;
         GLuint                     haxes_;

         // Flags.
         bool                       is_function3_;
         bool                       recompile_;
         bool                       lists_created_;
	 // Display options flag bits.
	 int                        mode_;

   };

}

template <int GW,int GH>
inline double graph::Graph3d<GW,GH>::getRangeX0() const
{
   return x0_;
}

template <int GW,int GH>
inline double graph::Graph3d<GW,GH>::getRangeX1() const
{
   return x1_;
}

template <int GW,int GH>
inline double graph::Graph3d<GW,GH>::getRangeY0() const
{
   return y0_;
}

template <int GW,int GH>
inline double graph::Graph3d<GW,GH>::getRangeY1() const
{
   return y1_;
}

template <int GW,int GH>
inline bool graph::Graph3d<GW,GH>::hasZ() const
{
   return is_function3_;
}

template <int GW,int GH>
inline int graph::Graph3d<GW,GH>::mode(int m) const
{
   return mode_ & m;
}

template <int GW,int GH>
inline void graph::Graph3d<GW,GH>::setRangeX(double x0,double x1)
{
   if ( x0 < x1 ) {
      x0_ = x0;
      x1_ = x1;
      initGraphValues();
   }
   else {
      throw illegal_argument_error();
   }
}

template <int GW,int GH>
inline void graph::Graph3d<GW,GH>::setRangeY(double y0,double y1)
{
   if ( y0 < y1 ) {
      y0_ = y0;
      y1_ = y1;
      initGraphValues();
   }
   else {
      throw illegal_argument_error();
   }
}

template <int GW,int GH>
inline void graph::Graph3d<GW,GH>::setZ(double z)
{
   if ( is_function3_ ) {
      ((Function3Wrapper<GW,GH> *)function_)->setZ(z);
      initGraphValues();
      recompile_ = true;
   }
}

#endif
