#ifndef _graph_runtime_h_
#define _graph_runtime_h_

// (C) by Thalion Graphic Solutions e.K. 2001
//
// Permissions to copy, modify and redistribute
// granted, according to the conditions defined
// in the Gnu Public License (GPL).
// Please read the file gpl.txt provided with this
// distribution.

#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>

// Select the engine to use.
// The instruction sets are not compatible!
#ifdef USE_ACC
// Use the accumulator variant of the stack engine.
# define INSTRUCTIONSET <graph/runtime_ins_acc.h>
#else
// Use the plain stack engine.
# define INSTRUCTIONSET <graph/runtime_ins_stack.h>
#endif

namespace graph {

   class Function;

   //
   // Bytecode interpreter.
   //
   class Runtime {

      public:

         // Instruction set.
         #include INSTRUCTIONSET         

         // Possible errors.
         class division_by_zero {};
	 class illegal_instruction {};
	 class overflow {};
         class underflow {};
      
         // An instruction: one byte for opcode,
         // 3 for optional parameter.
         struct Ins  {
	    Ins(ins_t ins=(ins_t)0,unsigned par=0);
	    Ins(const Ins&);
	    Ins(Ins&);
            Ins& operator=(const Ins&);
            const char *print(char *) const;
            unsigned ins_: 8;
            unsigned par_:24;
         };
	 
	 // A value.
	 union Value {
            double           d_; // A real number.
            const Function  *f_; // A function. Yes, it is actually used const.
            const Ins       *i_; // Saved IP. 
            long             l_; // A integer number.
            Value           *v_; // Saved frame.
         };

         // Creates a runtime environment.
         Runtime(size_t ssize = 1024,size_t dsize = 1024);
	 
	 // Resets the stack.
	 void reset();

         // Push argument onto the interpreter stack.
         void push(double);
         void push(const Function *);
         void push(int);
         void push(long);

         // Set a data segment entry.
	 // Returns true, if index was within bounds, else false.
         bool setData(unsigned,double);
         bool setData(unsigned,const Function&);
         bool setData(unsigned,long);

         // Evaluate a code piece.
         void main(const Ins *);

         // If a function returning real values was
         // previously invoked, this retrieves the result.
         double getRealResult();

         // If the function previously called, was to
         // return a function, this will return it.
         const Function *getFunctionResult();

         // For an integer function, this will return
         // the computation result.
         long getIntegerResult();
   
         // Destroy the runtime instance.
         virtual ~Runtime();

      private:
         Value  *ss_,      // The stack base.
                *sp_,      // The stack pointer.
                *stop_,    // Pointer to the stack top.
                *ds_;      // The data segment.
	 size_t  dsize_;   // Data segment max size.
   };

   //
   // A function executable by the interpreter.
   //
   class Function {

      public:

         // Create a function.
         Function(const char *,const char *,const Runtime::Ins *,size_t,int);

         // Get the name.
         const char *getName() const;

         // Get the type string.
         const char *getType() const;

         // Get the code segment.
         const Runtime::Ins *getCodeSegment() const;

         // Get the argument count.
         int getArgC() const;

         // Let it compute a real value.
         double computeReal(Runtime *,...) const;

         // Evaluate to function.
         const Function *computeFunction(Runtime *,...) const;

         // Evaluate to integer/boolean.
         long computeInteger(Runtime *,...) const;

         // Prints a function to a C file stream.
         void print(FILE *) const;

         // Prints a function to a stream.
         friend ostream& operator << (ostream&,const Function&);

         // Destruction.
         virtual ~Function();

      private:
         long           magic_;
         char          *name_;
         char          *type_;
         Runtime::Ins  *cs_;
         int            argc_;
   };

   // Prints an instruction as memnonic represenation
   // to a stream.
   ostream& operator << (ostream&,Runtime::Ins);

   // Prints a buffer of instructions in memnonic representation
   // to a stream.
   ostream& operator << (ostream&,const Runtime::Ins *);

}

inline void graph::Runtime::reset()
{
   sp_ = ss_ - 1;
}

inline void graph::Runtime::push(double d)
{
   (++sp_)->d_ = d; 
}

inline void graph::Runtime::push(const graph::Function *f)
{
   (++sp_)->f_ = f; 
}

inline void graph::Runtime::push(int i)
{
   (++sp_)->l_ = i; 
}

inline void graph::Runtime::push(long l)
{
   (++sp_)->l_ = l; 
}

inline bool graph::Runtime::setData(unsigned idx,double d)
{
   return ( idx < dsize_ ) ? ((ds_[idx].d_ = d), true) : false;
}

inline bool graph::Runtime::setData(unsigned idx,const graph::Function& f)
{
   return ( idx < dsize_ ) ? ((ds_[idx].f_ = &f), true) : false;
}

inline bool graph::Runtime::setData(unsigned idx,long l)
{
   return ( idx < dsize_ ) ? ((ds_[idx].l_ = l), true) : false;
}

inline double graph::Runtime::getRealResult()
{
   return ss_->d_;
}

inline const graph::Function *graph::Runtime::getFunctionResult()
{
   return ss_->f_;
}

inline long graph::Runtime::getIntegerResult()
{
   return ss_->l_;
}
   
inline const char *graph::Function::getName() const
{
   return name_;
}

inline const char *graph::Function::getType() const
{
   return type_;
}

inline const graph::Runtime::Ins *graph::Function::getCodeSegment() const
{
   return cs_;
}

inline int graph::Function::getArgC() const
{
   return argc_;
}

#endif
