package graph3d;

// (C) by Thalion Graphic Solutions e.K. 2001
//
// Permissions to copy, modify and redistribute
// granted, according to the conditions defined
// in the Gnu Public License (GPL).
// Please read the file gpl.txt provided with this
// distribution.

import graph3d.peer.CodeGen;
import graph3d.peer.DataBufferExceeded;

import mathc.FunctionDefinition;
import mathc.FunctionDescriptor;

import java.util.Vector;

public class CodeGenerator
   implements mathc.CodeGenerator {
   
   private static CodeGenerator instance_;
   
   static {
      instance_ = new CodeGenerator();
   }
   
   /**
    * Thrown, if the data buffer size required by this
    * code generator exceeds the jgraph stack machine data
    * buffer size.
    */
   public static class DataSizeExceededException
      extends mathc.CodeGenerator.Exception {
      public DataSizeExceededException()
      {
         super();
      }
   }
   
   static class Label {
      private int src_;
      private int target_;
      Label(int src,int target)
      {
         src_ = src;
	 target_ = target;
      }
      Label()
      {
         this(0,0);
      }
      int getSrc()
      {
         return src_;
      }
      int getTarget()
      {
         return target_;
      }
      void setSrc(int src)
      {
         src_ = src;
      }
      void setTarget(int target)
      {
         target_ = target;
      }
   }  

   private Vector labels_;
   private int    lastlabel_;
   private int    addr_;
   private int    hfunc_;

   private Vector functions_;
   
   CodeGenerator()
   {
      labels_    = null;
      lastlabel_ = 0;
      addr_      = -1;
      functions_ = new Vector();
   }

   /**
    * @return the CodeGenerator instance
    */
   public static CodeGenerator getInstance()
   {
      return instance_;
   }
   
   /**
    * Flushes the internal function list.
    */
   public void flush()
   {
      functions_.removeAllElements();
      CodeGen.reset();
      addr_ = -1;
      lastlabel_ = 0;
   } 

   /**
    * Returns a vector of functions registered by
    * the generator.
    * @return a vector containing all functions known
    *         by the code generator
    */
   public Vector getFunctions()
   {
      return functions_;
   }

   /**
    * The method creates a new function.
    * It returns a handle to this function.
    * @param name the function's name
    * @param proto the  function's prototype
    * @return a handle to the function
    */
   public int beginFunction(String name,
                            mathc.FunctionDefinition function)
   {
      String sig = ((FunctionDescriptor)function.getType()).mangledName();
      sig = sig.replace('L','(');
      sig = sig.replace('R',')');
      labels_    = new Vector();
      lastlabel_ = 0;
      addr_      = 0;
      hfunc_     = CodeGen.beginFunction(name,sig,function.getArgC());
      return hfunc_;
   }

   /**
    * The method indicates the end of a function
    * definition. Here, the code generator actually
    * will create the function's code.
    */
   public int endFunction()
      throws mathc.CodeGenerator.Exception
   {
      for ( int i=0; i<labels_.size(); i++ ) {
         Label label = (Label) labels_.elementAt(i);
	 CodeGen.setLabel(label.getSrc(),label.getTarget());	    
      }
      try {
         CodeGen.endFunction();
	 functions_.addElement(CodeGen.getFunction());
         return hfunc_;
      }
      catch ( DataBufferExceeded e ) {
         throw new DataSizeExceededException();
      }
   }

   /**
    * The method returns a handle to a new label.
    */
   public int newLabel()
   {
      Label label = new Label();
      labels_.addElement(label);
      return labels_.size() - 1;
   }

   /**
    * Sets a label.
    * @param hlabel a label handle
    */
   public void setLabel(int hlabel)
   {
      Label label = (Label)labels_.elementAt(hlabel);
      label.setTarget(addr_+1);
   }

   /** PUSHFC (push float constant) opcode.
       @param value a float constant */
   public void pushfc(double value)
      throws mathc.CodeGenerator.Exception
   {
      try {
         int hconst = CodeGen.constant((long)value);   
         addr_ = CodeGen.addIns(CodeGen.PUSHC,hconst);    
      }
      catch ( DataBufferExceeded e ) {
         throw new DataSizeExceededException();
      }
   }
   
   /** PUSHIC (push int constant) opcode.
       @param value an int value */
   public void pushic(int value)
      throws mathc.CodeGenerator.Exception
   {
      if ( (value << 8) >> 8 == value ) {
         addr_ = CodeGen.addIns(CodeGen.PUSHI,value);
      }
      else {
         try {
            int hconst = CodeGen.constant((long)value);
            addr_ = CodeGen.addIns(CodeGen.PUSHC,hconst);    
         }
         catch ( DataBufferExceeded e ) {
            throw new DataSizeExceededException();
         }
      }
   }
   
   /** PUSHXC (push executable) opcode.
       @param hconst a function handle */
   public void pushxc(int hconst)
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.PUSHC,hconst);    
   }
   
   /** PUSHFL (push float local var) opcode.
       @param hlocal a float local var address */
   public void pushfl(int hlocal)
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.PUSHL,hlocal);      
   }
   
   /** PUSHIL (push int local var) opcode.
       @param hlocal an int local var address */
   public void pushil(int hlocal)
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.PUSHL,hlocal);       
   }
   
   /** PUSHXL (push function local var) opcode.
       @param hlocal a function local var address */
   public void pushxl (int hlocal)
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.PUSHL,hlocal);       
   }
   
   /** PUSHT (push true) opcode. */
   public void pusht()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.PUSHI,1);     
   }
   
   /** PUSHF (push false) opcode. */
   public void pushf()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.PUSHI,0);        
   }
   
   /** STBL (store boolean local var) opcode.
       @param hlocal a boolean local var address */
   public void stbl(int hlocal)
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.STL,hlocal);        
   }
   
   /** STFL (store float local var) opcode.
       @param hlocal a float local var address */
   public void stfl(int hlocal)
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.STL,hlocal);          
   }
   
   /** STIL (store int local var) opcode.
       @param hlocal an int local var address */
   public void stil(int hlocal)
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.STL,hlocal);          
   }
   
   /** STXL (store function local var) opcode.
       @param hlocal a function local var address */
   public void stxl (int hlocal)
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.STL,hlocal);          
   }
   
   /** POP opcode. */
   public void pop()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.POP);     
   }
   
   /** POP2 opcode. */
   public void pop2()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.POP);     
   }
   
   /** DUP (duplicate item on stack top) opcode. */
   public void dup()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.DUP);     
   }
   
   /** DUP2 (duplicate item on stack top) opcode. */
   public void dup2()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.DUP);     
   }
   
   /** EXCH (exchange 2 1-cell items on opstack top) opcode. */
   public void exch_11()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.EXCH);     
   }
   
   /** EXCH (exchange 2 2-cell items on opstack top) opcode. */
   public void exch_22()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.EXCH);     
   }
   
   /** EXCH (exchange 1-cell and 2-cell item on opstack top) opcode. */
   public void exch_12()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.EXCH);     
   }
   
   /** EXCH (exchange 2-cell and 1-cell item on opstack top) opcode. */
   public void exch_21()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.EXCH);     
   }
   
   /** CALL (call function) opcode. */
   public void call(FunctionDescriptor proto)
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.CALL);     
   }
   
   /** BRET (return boolean) opcode. */
   public void bret()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.RET);     
   }
   
   /** FRET (return float) opcode. */
   public void fret()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.RET);     
   }
   
   /** IRET (return integer) opcode. */
   public void iret()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.RET);     
   }
   
   /** XRET (return function) opcode. */
   public void xret()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.RET);       
   }
   
   /** J (uncoditional jump) opcode.
       @param hlabel a handle to the jump target label */
   public void j(int hlabel)
      throws mathc.CodeGenerator.Exception
   {
      Label label = (Label)labels_.elementAt(hlabel);   
      addr_ = CodeGen.addIns(CodeGen.J);
      label.setSrc(addr_);
   }
   
   /** J (jump if true) opcode.
       @param hlabel a handle to the jump target label */
   public void jt(int hlabel)
      throws mathc.CodeGenerator.Exception
   {
      Label label = (Label)labels_.elementAt(hlabel);
      addr_ = CodeGen.addIns(CodeGen.JT);
      label.setSrc(addr_);      
   }
   
   /** J (jump if false) opcode.
       @param hlabel a handle to the jump target label */
   public void jf(int hlabel)
      throws mathc.CodeGenerator.Exception
   {
      Label label = (Label)labels_.elementAt(hlabel);
      addr_ = CodeGen.addIns(CodeGen.JF);
      label.setSrc(addr_);      
   }
   
   /** B2J (convert bool to int) opcode. */
   public void b2i()
      throws mathc.CodeGenerator.Exception
   {
      // Nothing to do.
   }
   
   /** F2I (convert float to int) opcode. */
   public void f2i()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.F2I);
   }
   /** I2B (convert int to bool) opcode. */
   public void i2b()
      throws mathc.CodeGenerator.Exception
   {
      // Nothing to do.
   }
    
   /** I2F (convert int to float) opcode. */
   public void i2f()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.I2F);
   }
   
   /** BEQ (boolean equality) opcode. */
   public void beq()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.EQ);   
   }
   
   /** BNE (boolean not equal) opcode. */
   public void bne()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.NE);    
   }
   
   /** FADD (float add) opcode. */
   public void fadd()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.FADD);       
   }
   
   /** FSUB (float subtract) opcode. */
   public void fsub()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.FSUB);       
   }
   
   /** FMUL (float multiply) opcode. */
   public void fmul()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.FMUL);        
   }
   
   /** FDIV (float divide) opcode. */
   public void fdiv()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.FDIV);        
   }
   
   /** FPOW (float power) opcode. */
   public void fpow()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.FPOW);    
   }
   
   /** FNEG (float negate) opcode. */
   public void fneg()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.FNEG);     
   }
   
   /** FGT (float greater than) opcode. */
   public void fgt()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.FGT);       
   }
   
   /** FGE (float greater or equals) opcode. */
   public void fge()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.FGE);       
   }
   
   /** FLT (float less than) opcode. */
   public void flt()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.FLT);     
   }
   
   /** FLE (float less or equals) opcode. */
   public void fle()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.FLE);     
   }
   
   /** FEQ (float equality) opcode. */
   public void feq()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.EQ);     
   }
   
   /** FNE (float inequality) opcode. */
   public void fne()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.NE);     
   }
   
   /** IADD (integer add) opcode. */
   public void iadd()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.IADD);       
   }
   
   /** ISUB (integer subtract) opcode. */
   public void isub()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.ISUB);       
   }
   
   /** IMUL (integer multiply) opcode. */
   public void imul()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.IMUL);       
   }
   
   /** IDIV (integer divide) opcode. */
   public void idiv()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.IDIV);       
   }
   
   /** IMOD (modulo) opcode. */
   public void imod()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.IMOD);       
   }
   
   /** IPOW (integer power) opcode. */
   public void ipow()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.IPOW);       
   }
   
   /** INEG (integer negate) opcode. */
   public void ineg()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.INEG);       
   }
   
   /** IGT (integer greater than) opcode. */
   public void igt()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.IGT);       
   }
   
   /** IGE (integer greater or equals) opcode. */
   public void ige()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.IGE);       
   }
   
   /** ILT (integer less than) opcode. */
   public void ilt()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.ILT);       
   }
   
   /** ILE (integer less or equals) opcode. */
   public void ile()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.ILE);       
   }
   
   /** IEQ (integer equality) opcode. */
   public void ieq()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.EQ);       
   }
   
   /** INE (integer not equal) opcode. */
   public void ine()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.NE);       
   }
   
   /** XEQ (function equality) opcode. */
   public void xeq()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.EQ);       
   }
   
   /** XNE (function not equal) opcode. */
   public void xne()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.NE);       
   }
   
   /** AND (logical and) opcode. */
   public void and()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.AND);       
   }
   
   /** OR (logical or) opcode. */
   public void or()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.OR);       
   }
   
   /** XOR (logical exclusive or) opcode. */
   public void xor()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.XOR);       
   }
   
   /** NOT (logical negate) opcode. */
   public void not()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.NOT);      
   }
   
   /** SIN (trignometrical sine function) opcode. */
   public void sin()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.SIN);     
   }
   
   /** COS (trigonometrical cosine function) opcode. */
   public void cos()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.COS);     
   }
   
   /** TAN (trigonometrical tan function) opcode. */
   public void tan()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.TAN);     
   }
   
   /** ASIN (trigonometrical asine function) opcode. */
   public void asin()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.ASIN);     
   }
   
   /** ACOS (trigonometrical acosine function) opcode. */
   public void acos()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.ACOS);     
   }
   
   /** ATAN (trigonometrical atan function) opcode. */
   public void atan()
      throws mathc.CodeGenerator.Exception
   {
      addr_ = CodeGen.addIns(CodeGen.ATAN);     
   }

   
}
   
