package graph3d;

// (C) by Thalion Graphic Solutions e.K. 2001
//
// Permissions to copy, modify and redistribute
// granted, according to the conditions defined
// in the Gnu Public License (GPL).
// Please read the file gpl.txt provided with this
// distribution.

/**
 * AWT imports.
 */
import java.awt.EventQueue;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Image;
import java.awt.PrintJob;
import java.awt.Toolkit;

import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import java.awt.image.ImageObserver;
import java.awt.image.MemoryImageSource;

import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;

/**
 * Local imports.
 */
import graph3d.peer.Graph3d;

/**
 * A graph view.
 * This class is threadsafe, ie. one can send all kind
 * of state setter methods in the context of any thread.
 * Note: this explicitly excludes paint() and listener methods,
 *       which are not meant to be invoked immediately anyway.
 */
public class GraphView
   extends Canvas
   implements ComponentListener,
	      MouseMotionListener,
              Printable {
	      
   private static abstract class QueuedAction
      implements Runnable {
      private Canvas c_;
      public QueuedAction(Canvas c)
      {
         c_ = c;
      }
      protected Canvas getCanvas()
      {
         return c_;
      }
   }
   
   /**
    * Mode bit: indicates graph shading as point set.
    */
   public static final int POINTS   = Graph3d.POINTS;

   /**
    * Mode bit: indicates graph shading as lines.
    */
   public static final int LINES    = Graph3d.LINES;

   /**
    * Mode bit: indicates smooth shading of graph.
    */
   public static final int SMOOTH   = Graph3d.SMOOTH;
   
   /**
    * Mode bit: indicates usage of lighting.
    */
   public static final int LIGHTS   = Graph3d.LIGHTS;
   
   /**
    * Mode bit: indicates usage of x,y,z axis display.
    */
   public static final int AXES     = Graph3d.AXES;
   
   /**
    * Mode bit: draw contour colors.
    */
   public static final int CONTOURS = Graph3d.CONTOURS;

   /**
    * Mode bit: use display lists.
    */
   public static final int DLISTS = Graph3d.DLISTS;
	
   /**	     
    * The configuration to use for the canvas.
    */
   private static java.awt.GraphicsConfiguration vp_configuration_;
   
   /**	     
    * The configuration to use for offscreen drawing.
    */
   private static java.awt.GraphicsConfiguration offs_configuration_;

   /**
    * The cursor.
    */
   private static Cursor cursor_;
     
   /**
    * Find the best graphics configuration here,
    * and initialize handles.
    */
   static {
      GraphicsEnvironment env = GraphicsEnvironment.getLocalGraphicsEnvironment();
      GraphicsDevice dev = env.getDefaultScreenDevice();
      vp_configuration_ =
         dev.getBestConfiguration(new GraphicsConfigTemplate(GraphicsConfigTemplate.VIEWPORT));
      offs_configuration_ =
         dev.getBestConfiguration(new GraphicsConfigTemplate(GraphicsConfigTemplate.OFFSCREEN));
      cursor_ = new Cursor(Cursor.CROSSHAIR_CURSOR);
   }

   /**
    * The peer of this class.
    */
   private graph3d.peer.Graph3d graph_;
   
   /**
    * Is the graph already realized?
    */
   private boolean realized_ = false;
   
   /**
    * Recalled mouse drag screen coordinates.
    */
   private int old_x_, old_y_;
   
   /**
    * Current rotation angles around x,y,z axis.
    */
   private double ax_, ay_, az_;

   /**
    * Create a graph from a function.
    * @param function the function to graph
    * @param x0 low bound of x range
    * @param x1 high bound of x range
    * @param y0 low bound of y range
    * @param y1 high bound of y range
    * @param mode the mode bits to use  
    * @throws IllegalArgumentException, if the function is not plottable
    *         RuntimeException, if the function causes a stack machine error
    */
   public GraphView(Function function,
                    double x0,
                    double x1,
                    double y0,
                    double y1,
                    int mode)
      throws IllegalArgumentException,
             RuntimeException,
             TypeMismatchException
   {
      super(vp_configuration_);
      graph_ = new Graph3d(function,x0,x1,y0,y1,mode);
      addComponentListener(this);
      addMouseMotionListener(this);
      setCursor(cursor_);
   }

   /**
    * Create a graph from a function.
    * @param function the function to graph
    * @throws IllegalArgumentException, if the function is not plottable
    *         RuntimeException, if the function causes a stack machine error
    */
   public GraphView(Function function)
      throws IllegalArgumentException,
             RuntimeException,
             TypeMismatchException
   {
      super(vp_configuration_);
      graph_ = new Graph3d(function);
      addComponentListener(this);
      addMouseMotionListener(this);
      setCursor(cursor_);
   }

   /**
    * This method initializes the graph's color parameters to non-default
    * values. It should be invoked before the first paint().
    * Use the setBackgroundColor() and setLightColor() methods to alter
    * color parameters later.
    * @param l0r light #0 red component
    * @param l0g light #0 green component
    * @param l0b light #0 blue component
    * @param l1r light #1 red component
    * @param l1g light #1 green component
    * @param l1b light #1 blue component
    * @param bgr background/clear color red component
    * @param bgg background/clear color green component
    * @param bgb background/clear color blue component
    */
   public void initColors(float l0r,float l0g,float l0b,
                          float l1r,float l1g,float l1b,
                          float bgr,float bgg,float bgb)
   {
      graph_.initColors(l0r,l0g,l0b,l1r,l1g,l1b,bgr,bgg,bgb);
   }

   /**
    * This method initializes the graph's color parameters to non-default
    * values. It should be invoked before the first paint().
    * Use the setBackgroundColor() and setLightColor() methods to alter
    * color parameters later.
    * @param l0 light #0 color
    * @param l1 light #1 color
    * @param bg background/clear color
    */
   public void initColors(Color l0,Color l1,Color bg)
   {
      initColors(((float)l0.getRed())/255,
                 ((float)l0.getGreen())/255,
                 ((float)l0.getBlue())/255,
                 ((float)l1.getRed())/255,
                 ((float)l1.getGreen())/255,
                 ((float)l1.getBlue())/255,
                 ((float)bg.getRed())/255,
                 ((float)bg.getGreen())/255,
                 ((float)bg.getBlue())/255);
   }
   
   /**
    * Not used.
    */
   public void componentHidden(ComponentEvent event)
   {
      // Empty.
   }
   
   /**
    * Not used.
    */
   public void componentMoved(ComponentEvent event)
   {
      // Empty.
   }
   
   /**
    * Invokes a reshape.
    * This method must not be called from outside the event thread.
    */
   public void componentResized(ComponentEvent event)
   {
      if ( ! EventQueue.isDispatchThread() ) {
         throw new IllegalArgumentException("not in dispatch thread");
      }
      if ( realized_ ) {
         graph_.reshape(this,getWidth(),getHeight());
      }
   }

   /**
    * Not used.
    */
   public void componentShown(ComponentEvent event)
   {
      // Empty. 
   } 

   /**
    * Not used.
    */
   public void mouseClicked(MouseEvent event)
   {
      // Empty.
   }
   
   /**
    * Reacts on mouse drags.
    * Used to compute scene rotation.
    */
   public void mouseDragged(MouseEvent event)
   {
      if ( ! EventQueue.isDispatchThread() ) {
         throw new IllegalArgumentException("not in dispatch thread");
      }   
      if ( (event.getModifiers() & MouseEvent.BUTTON1_MASK) != 0 ) {
         screenRotate(old_x_,old_y_,event.getX(),event.getY());
         old_x_ = event.getX();
         old_y_ = event.getY();
         return;
      }
      if ( (event.getModifiers() & MouseEvent.BUTTON3_MASK) != 0 ) {
         zoom( (event.getX() > old_x_ ) ? 0.1 : -0.1);
         old_x_ = event.getX();
         return;
      }
   }

   /**
    * Not used.
    */
   public void mouseMoved(MouseEvent event)
   {
      // Empty.
   }
   
   /**
    * Paint the view.
    * This method must not be called from outside the event thread.
    * @param g the Graphics object
    * @throws IllegalArgumentException, if invoked from another thread
    */
   public void paint(Graphics g)
   {
      if ( ! EventQueue.isDispatchThread() ) {
         throw new IllegalArgumentException("not in dispatch thread");
      }
      if ( ! realized_ ) {
         graph_.realize(this);
	 realized_ = true;
      }
      graph_.display(this);
   }

   /**
    * This method prints this view in a print job's context.
    * @param job the print job to use
    */
   public int print(Graphics g,PageFormat page,int page_index)
      throws PrinterException
   {
      if ( page_index == 0 ) {
         //int width = (int) page.getWidth();
         //int height = (int) page.getHeight();
         int width = (int) (page.getWidth() / 72 ) * 150;
         int height = (int) (page.getHeight() / 72) * 150;
         int[] pixels =
            graph_.getImageData(this,width,height);
         MemoryImageSource source =
            new MemoryImageSource(width,height,offs_configuration_.getColorModel(),pixels,0,width);
         Image image = createImage(source);
         g.drawImage(image,0,0,(int)page.getWidth(),(int)page.getHeight(),new ImageObserver() {
            public boolean imageUpdate(Image img,
                                       int infoflags,
                                       int x,
                                       int y,
                                       int width,
                                       int height)
            {
               System.out.println("x = " + x + ", y = " + y + ", w = " + width + ", h = " + height);
               return true;
            }
         });
         return PAGE_EXISTS;
      }
      else {
         return NO_SUCH_PAGE;
      }
   }
   
   
   /**
    * Returns the red component of light #n color.
    * @return red component of color of light #n
    * @throws IllegalArgumentException, if n does not match a valid light
    */
   public float getLightRedValue(int n)
   {
      return graph_.getLightRedValue(n);
   }
   
   /**
    * Returns the green component of light #n color.
    * @return green component of color of light #n
    * @throws IllegalArgumentException, if n does not match a valid light
    */
   public float getLightGreenValue(int n)
   {
      return graph_.getLightGreenValue(n);
   }
   
   /**
    * Returns the blue component of light #n color.
    * @return blue component of color of light #n
    * @throws IllegalArgumentException, if n does not match a valid light
    */
   public float getLightBlueValue(int n)
   {
      return graph_.getLightBlueValue(n);
   }
   
   /**
    * Returns the X range low boundary.
    * @return low boundary on X axis
    */
   public double getRangeX0()
   {
      return graph_.getRangeX0();
   }

   /**
    * Returns the X range high boundary.
    * @return high boundary on X axis
    */
   public double getRangeX1()
   {
      return graph_.getRangeX1();
   }

   /**
    * Returns the Y range low boundary.
    * @return low boundary on Y axis
    */
   public double getRangeY0()
   {
      return graph_.getRangeY0();
   }

   /**
    * Returns the Y range high boundary.
    * @return high boundary on Y axis
    */
   public double getRangeY1()
   {
      return graph_.getRangeY1();
   }
   
   /**
    * Checks, if this is a f(x,y,z)-style function.
    * @return true, if this graph allows z value setting
    */
   public boolean hasZ()
   {
      return graph_.hasZ();
   } 
   
   /**
    * Checks mode bits.
    * @param m the mode bit(s) to check for.
    * @return the mode bits matched
    */
   public int mode(int m)
   {
      return graph_.mode(m);
   }
   
   /**
    * Rotate scene around X axis.
    * @param ax angle to apply
    */  
   public void rotateX(final double ax)
   {
      ax_ += ax;
      if ( realized_ && isShowing() ) {
         if ( EventQueue.isDispatchThread() ) {
            graph_.rotateX(this,ax_);
	    graph_.display(this);
         }
         else {
            QueuedAction a = new QueuedAction(this) {
	       public void run()
	       {
	          graph_.rotateX(getCanvas(),ax_);
		  graph_.display(getCanvas());
	       }
	    };
	    EventQueue.invokeLater(a);
         }
      }
   }
   
   /**
    * Rotate scene around Y axis.
    * @param ay angle to apply
    */  
   public void rotateY(final double ay)
   {
      ay_ += ay;
      if ( realized_ && isShowing() ) {
         if ( EventQueue.isDispatchThread() ) {
            graph_.rotateY(this,ay_);
	    graph_.display(this);
         }
         else {
            QueuedAction a = new QueuedAction(this) {
	       public void run()
	       {
	          graph_.rotateY(getCanvas(),ay_);
		  graph_.display(getCanvas());
	       }
	    };
	    EventQueue.invokeLater(a);
         }
      }
   }
   
   /**
    * Rotate scene around Z axis.
    * @param az angle to apply
    */ 
   public void rotateZ(final double az)
   {
      az_ += az;
      if ( realized_ && isShowing() ) {
         if ( EventQueue.isDispatchThread() ) {
            graph_.rotateZ(this,az_);
	    graph_.display(this);
         }
         else {
            QueuedAction a = new QueuedAction(this) {
	       public void run()
	       {
	          graph_.rotateZ(getCanvas(),az_);
		  graph_.display(getCanvas());
	       }
	    };
	    EventQueue.invokeLater(a);
         }
      }
   }
   
   /**
    * Rotate around axis given by cross product of reverse projected
    * screen coordinate vectors.
    * @param flag - lighting is switched on, if true, else switched off
    */
   private void screenRotate(final int x0,final int y0,final int x1,final int y1)
   {
      /*ax_ += Math.PI * (y1 - y0) / (double) getHeight();
      ay_ += Math.PI * (x1 - x0) / (double) getWidth();
      graph_.rotateX(this,ax_);
      graph_.rotateY(this,ay_);
      graph_.display(this);*/
      graph_.screenRotate(this,x0,y0,x1,y1);
      graph_.display(this);
   }

   /**
    * Set background color of view.
    * @param r red value of color
    * @param g green value of color
    * @param b blue value of color
    */
   public void setBackground(final float r, final float g,final float b)
   {
      if ( realized_ && isShowing() ) {
         if ( EventQueue.isDispatchThread() ) {
            graph_.setBackgroundColor(this,r,g,b);
	    graph_.display(this);
         }
         else {
            QueuedAction a = new QueuedAction(this) {
	       public void run()
	       {
	          graph_.setBackgroundColor(getCanvas(),r,g,b);
		  graph_.display(getCanvas());
	       }
	    };
	    try {
	       EventQueue.invokeAndWait(a);
	    }
	    catch ( IllegalArgumentException e ) {
	       throw e;
	    }
	    catch ( Exception e ) {
	       e.printStackTrace();
	    }
         }
      }
   }

   /**
    * Set background color of view.
    * @param c the color
    */
   public void setBackground(Color c)
   {
      setBackground(((float)c.getRed())/255,
                    ((float)c.getGreen())/255,
                    ((float)c.getBlue())/255);
   }
   
   /**
    * Set light #n color.
    * @param n number of light
    * @param r red value of color
    * @param g green value of color
    * @param b blue value of color
    * @throws IllegalArgumentException, if n is not a valid light
    */
   public void setLightColor(final int n, final float r, final float g,final float b)
      throws IllegalArgumentException
   {
      if ( realized_ && isShowing() ) {
         if ( EventQueue.isDispatchThread() ) {
            graph_.setLightColor(this,n,r,g,b);
	    graph_.display(this);
         }
         else {
            QueuedAction a = new QueuedAction(this) {
	       public void run()
	       {
	          graph_.setLightColor(getCanvas(),n,r,g,b);
		  graph_.display(getCanvas());
	       }
	    };
	    try {
	       EventQueue.invokeAndWait(a);
	    }
	    catch ( IllegalArgumentException e ) {
	       throw e;
	    }
	    catch ( Exception e ) {
	       e.printStackTrace();
	    }
         }
      }
   }

   /**
    * Set light #n color.
    * @param n number of light
    * @param c the color instance
    * @throws IllegalArgumentException, if n is not a valid light
    */
   public void setLightColor(int n,Color c)
      throws IllegalArgumentException
   {
      setLightColor(n,
                    ((float)c.getRed())/255,
                    ((float)c.getGreen())/255,
                    ((float)c.getBlue())/255);
   }
   
   /**
    * Set default rendering mode.
    * Only the POINTS flag will be set, and the corresponding
    * drawing mode be used.
    */
   public void setDefaultMode(final int mode)
   {
      if ( realized_ && isShowing() ) {
         if ( EventQueue.isDispatchThread() ) {
            graph_.setDefaultMode(this);
	    graph_.display(this);
         }
         else {
            QueuedAction a = new QueuedAction(this) {
	       public void run()
	       {
	          graph_.setDefaultMode(getCanvas());
		  graph_.display(getCanvas());
	       }
	    };
	    EventQueue.invokeLater(a);
         }
      }
   }

   /**
    * Set rendering mode.
    * @param mode POINTS: draws the grid as a point set
    *             LINES : draws the grid as a wireframe mesh
    *             SMOOTH: applies smooth shading on the grid
    *             LIGHTS: uses lighting
    *             AXES  : draws x,y,z axes
    */
   public void setMode(final int mode)
   {
      if ( realized_ && isShowing() ) {
         if ( EventQueue.isDispatchThread() ) {
            graph_.setMode(this,mode);
	    graph_.display(this);
         }
         else {
            QueuedAction a = new QueuedAction(this) {
	       public void run()
	       {
	          graph_.setMode(getCanvas(),mode);
		  graph_.display(getCanvas());
	       }
	    };
	    EventQueue.invokeLater(a);
         }
      }
   }
   
   /**
    * Set plot range of X axis.
    * @param x0 low bound of X range
    * @param x1 high bound of X range
    * @throws IllegalArgumentException, if y0 >= y1
    */
   public void setRangeX(final double x0,final double x1)
      throws IllegalArgumentException
   {
      if ( realized_ && isShowing() ) {
         if ( EventQueue.isDispatchThread() ) {
            graph_.setRangeX(this,x0,x1);
	    graph_.display(this);
         }
         else {
            QueuedAction a = new QueuedAction(this) {
	       public void run()
	       {
	          graph_.setRangeX(getCanvas(),x0,x1);
		  graph_.display(getCanvas());
	       }
	    };
	    try {
	       EventQueue.invokeAndWait(a);
	    }
	    catch ( IllegalArgumentException e ) {
	       throw e;
	    }
	    catch ( Exception e ) {
	       e.printStackTrace();
	    }
         }
      }
   }
   
   /**
    * Set plot range of Y axis.
    * @param y0 low bound of Y range
    * @param y1 high bound of Y range
    * @throws IllegalArgumentException, if y0 >= y1
    */   
   public void setRangeY(final double y0,final double y1)
      throws IllegalArgumentException
   {
      if ( realized_ && isShowing() ) {
         if ( EventQueue.isDispatchThread() ) {
            graph_.setRangeY(this,y0,y1);
	    graph_.display(this);
         }
         else {
            QueuedAction a = new QueuedAction(this) {
	       public void run()
	       {
	          graph_.setRangeY(getCanvas(),y0,y1);
		  graph_.display(getCanvas());
	       }
	    };
	    try {
	       EventQueue.invokeAndWait(a);
	    }
	    catch ( IllegalArgumentException e ) {
	       throw e;
	    }
	    catch ( Exception e ) {
	       e.printStackTrace();
	    }
         }
      }	 
   }
   
   /**
    * Set Z value for a f(x,y,z)-style function.
    * This can be used to animate a plot.
    * @param z Z value to use for f(x,y,z)
    */ 
   public void setZ(final double z)
   {
      if ( realized_ && isShowing() ) {
         if ( EventQueue.isDispatchThread() ) {
            graph_.setZ(z);
	    graph_.display(this);
         }
         else {
            QueuedAction a = new QueuedAction(this) {
	       public void run()
	       {
	          graph_.setZ(z);
		  graph_.display(getCanvas());
	       }
	    };
	    EventQueue.invokeLater(a);
         }
      }
   }
   
   /**
    * Unset a rendering mode bit.
    * @param mode POINTS: draws the grid as a point set
    *             LINES : draws the grid as a wireframe mesh
    *             SMOOTH: applies smooth shading on the grid
    *             LIGHTS: uses lighting
    *             AXES  : draws x,y,z axes
    */
   public void unsetMode(final int mode)
   {
      if ( realized_ && isShowing() ) {
         if ( EventQueue.isDispatchThread() ) {
            graph_.unsetMode(this,mode);
	    graph_.display(this);
         }
         else {
            QueuedAction a = new QueuedAction(this) {
	       public void run()
	       {
	          graph_.unsetMode(getCanvas(),mode);
		  graph_.display(getCanvas());
	       }
	    };
	    EventQueue.invokeLater(a);
         }
      }
   }
   
   /**
    * Zooms the graph in or out.
    * @param tz - values larger 0 zoom out
    *             values less than 0 zoom in
    */ 
   public void zoom(final double tz)
   {
      if ( realized_ && isShowing() ) {
         if ( EventQueue.isDispatchThread() ) {
            graph_.zoom(this,tz);
	    graph_.display(this);
         }
         else {
            QueuedAction a = new QueuedAction(this) {
	       public void run()
	       {
	          graph_.zoom(getCanvas(),tz);
		  graph_.display(getCanvas());
	       }
	    };
	    EventQueue.invokeLater(a);
         }
      }
   }
   
   public void finalize()
   {
      if ( graph_ != null ) {
         graph_.dispose(this);
         graph_ = null;
      }
   }

}
