package graph3d.peer;

// (C) by Thalion Graphic Solutions e.K. 2001
//
// Permissions to copy, modify and redistribute
// granted, according to the conditions defined
// in the Gnu Public License (GPL).
// Please read the file gpl.txt provided with this
// distribution.

import graph3d.Function;

/**
 * The code generator peer.
 * As there is only one generator per jgraph context,
 * this class does only have static members.
 */
public class CodeGen {

    /**
     * Below, opcodes understood by jgraph's internal interpreter,
     * are defined as memnonic constants.
     *
     * This is the instruction set used by the pure stack based vm.
     * Don't run this class with shared libraries built with the acc vm;
     * it uses a slightly different instruction set.
     */

    public static final int PUSHC      =  0; // Constant from DS to stack
    public static final int PUSHI      =  1; // Immediate to stack
    public static final int PUSHL      =  2; // Local to stack
    public static final int STL        =  3; // Acc to local
    public static final int POP        =  4;
    public static final int DUP        =  5;
    public static final int EXCH       =  6;
    public static final int CALL       =  7;
    public static final int RET        =  8;
    public static final int J          =  9;
    public static final int JT         = 10;
    public static final int JF         = 11;
    public static final int F2I        = 12;
    public static final int I2F        = 13;
    public static final int FADD       = 14;
    public static final int FSUB       = 15;
    public static final int FMUL       = 16;
    public static final int FDIV       = 17;
    public static final int FPOW       = 18;
    public static final int FNEG       = 19;
    public static final int IADD       = 20;
    public static final int ISUB       = 21;
    public static final int IMUL       = 22;
    public static final int IDIV       = 23;
    public static final int IMOD       = 24;
    public static final int IPOW       = 25;
    public static final int INEG       = 26;
    public static final int EQ         = 27;
    public static final int NE         = 28;
    public static final int IGT        = 29;
    public static final int IGE        = 30;
    public static final int ILT        = 31;
    public static final int ILE        = 32;
    public static final int FGT        = 33;
    public static final int FGE        = 34;
    public static final int FLT        = 35;
    public static final int FLE        = 36;
    public static final int AND        = 37;
    public static final int OR         = 38;
    public static final int XOR        = 39;
    public static final int NOT        = 40;
    public static final int SIN        = 41;
    public static final int COS        = 42;
    public static final int TAN        = 43;
    public static final int ASIN       = 44;
    public static final int ACOS       = 45;
    public static final int ATAN       = 46;
    public static final int EOC        = 63;   // End Of Code mark.

   public static int beginFunction(String name,String signature,int argc)
   {
      return Context.getInstance().beginFunction(name,signature,argc);
   }   
   
   public static int addIns(int ins)
   {
      return Context.getInstance().addIns(ins);   
   }
   
   public static int addIns(int ins,int argc)
   {
      return Context.getInstance().addIns(ins,argc);     
   }
   
   public static void setLabel(int addr,int label)
   {
      Context.getInstance().setLabel(addr,label);
   }

   /**
    * Ends a function definition and tries to register
    * it in the graph3 runtime system. If there is not enough
    * free buffer space left, a DataBufferExceeded exception is
    * thrown.
    * @throws DataBufferExceeded, if there is no free buffer
    *         space left for this function
    */
   public static void endFunction()
      throws DataBufferExceeded
   {
      Context.getInstance().endFunction();     
   }
   
   /**
    * Resets the code generator.
    */
   public static void reset()
   {
      Context.getInstance().reset();
   }
   
   /**
    * Returns the current function.
    * If invoked, before endFunction() has been called, this method
    * will return null, else the new function instance.
    * @returns the new function instance
    */
   public static Function getFunction()
   {
      return new InterpretedFunction(Context.getInstance().getFunction());
   }
   
   /**
    * Registers a real number (double) constant and returns it's handle.
    * Fails, if there is not enough free data buffer space left in
    * the graph3d interpreter system.
    * @param c the double constant
    * @return the data buffer handle of c
    * @throws DataBufferExceeded, if entry fails due to insufficent free
    *         buffer space
    */
   public static int constant(double c)
      throws DataBufferExceeded
   {
      return Context.getInstance().constant(c);
   }
   
   /**
    * Registers a integer (long) constant and returns it's handle.
    * Fails, if there is not enough free data buffer space left in
    * the graph3d interpreter system.
    * @param c the long constant
    * @return the data buffer handle of c
    * @throws DataBufferExceeded, if entry fails due to insufficent free
    *         buffer space
    */
   public static int constant(long c)
      throws DataBufferExceeded
   {
      return Context.getInstance().constant(c);
   }   

}
