package graph3d.peer;

// (C) by Thalion Graphic Solutions e.K. 2001
//
// Permissions to copy, modify and redistribute
// granted, according to the conditions defined
// in the Gnu Public License (GPL).
// Please read the file gpl.txt provided with this
// distribution.

import graph3d.Function;
import graph3d.RuntimeException;
import graph3d.TypeMismatchException;
import java.awt.Canvas;
import java.awt.GraphicsConfiguration;

/**
 * This class implements a globally synchronized interface
 * to the native library.
 * It's implemented using the singleton pattern.
 * All of it's method are marked up synchronized.
 * They're not meant to be immediately called, but via instances
 * of graph3d.peer.Graph3d and graph3d.peer.Function, which know about
 * their handles.
 */
final class Context {

   /**
    * The library context instance.
    */
   private static Context instance_ = null;

   /**
    * Initialize the library.
    * Create the context instance.
    */
   static {
      System.loadLibrary("jgraph");
      initLibrary();
      instance_ = new Context();
      instance_.findConfiguration();
   }
   
   /**
    * Initializes the native library.
    */
   private static native void initLibrary();
   
   /**
    * Returns the context.
    * @return the context
    */
   static Context getInstance()
   {
      return instance_;
   }
   
   /**
    * Checks, if the passed configuration matches the native 
    * rendering requirements.
    * @param gc the graphics configuration
    * @return true, if the configuration matches, else false
    */
   synchronized native boolean checkOffsConfiguration(GraphicsConfiguration gc);

   /**
    * Checks, if the passed configuration matches the native 
    * rendering requirements.
    * @param gc the graphics configuration
    * @return true, if the configuration matches, else false
    */
   synchronized native boolean checkVpConfiguration(GraphicsConfiguration gc);
   
   /**
    * Finds the graphics configuration settings required
    * for drawing.
    */
   synchronized native void findConfiguration();

   /**
    * Create a native object instance.
    * @param f the function to display
    * @param x0 low bound of x range
    * @param x1 high bound of x range
    * @param y0 low bound of y range
    * @param y1 high bound of y range
    * @param mode the render mode to use
    * @return a handle to the new native instance.
    * @throws IllegalArgumentException, if the x or y range is invalid
    * @throws RuntimeException, if the plotted function causes some runtime
    *         error on the internal stack machine
    * @throws TypeMismatchException, if the function prototype does not
    *         match one of the plottable types ( (ff)f and (fff)f currently)
    */
   synchronized native long create(Function f,
                                   double x0,
                                   double x1,
                                   double y0,
                                   double y1,
                                   int mode)
      throws IllegalArgumentException,
             RuntimeException,
             TypeMismatchException;

   /**
    * This method initializes the graph's color parameters to non-default
    * values. It should be invoked before the graph is been realized.
    * Use the setBackgroundColor() and setLightColor() methods to alter
    * color parameters later.
    * @param hgraph handle to the graph native instance
    * @param l0r light #0 red component
    * @param l0g light #0 green component
    * @param l0b light #0 blue component
    * @param l1r light #1 red component
    * @param l1g light #1 green component
    * @param l1b light #1 blue component
    * @param bgr background/clear color red component
    * @param bgg background/clear color green component
    * @param bgb background/clear color blue component
    */
   synchronized native void initColors(long hgraph,
                                       float l0r,float l0g,float l0b,
                                       float l1r,float l1g,float l1b,
                                       float bgr,float bgg,float bgb);
   
   /**
    * This method realizes a graph instance, ie. it actually creates
    * graphics data structures from the function data grid.
    * @param hgraph handle to the graph native instance
    * @param c the canvas this graph belongs to
    */
   synchronized native void realize(long hgraph,Canvas c);

   /**
    * Display the graph.
    * This method will (and should) be invoked from
    * within a canvas' paint() method. It may not be called
    * from outside the AWT event dispatch thread.
    * @param hgraph handle to the graph native instance
    * @param c the canvas to draw into
    */
   synchronized native void display(long hgraph,Canvas c);

   /**
    * Reshape the graph.
    * This method should be invoked, whenever a viewport resize
    * has occurred.
    * @param hgraph handle to the graph native instance
    * @param c the canvas which owns this graph
    * @param w the new viewport width in screen coordinates
    * @param h the new viewport height in screen coordinates
    */
   synchronized native void reshape(long hgraph,Canvas c,int w,int h);
   
   /**
    * Prints a frame into an off-screen buffer.
    * @param hgraph handle to the graph native instance
    * @param c the canvas this graph belongs to
    * @param w width of printable area
    * @param h height of printable area
    * @return an array of ints representing the pixels
    */
   synchronized native int[] getImageData(long hgraph,Canvas c,int w,int h);

   /**
    * Returns the red component of light #n color.
    * @param hgraph handle to the graph native instance
    * @return red component of color of light #n
    * @throws IllegalArgumentException, if n does not match a valid light
    */
   synchronized native float getLightRedValue(long hgraph,int n);
   
   /**
    * Returns the green component of light #n color.
    * @param hgraph handle to the graph native instance
    * @return green component of color of light #n
    * @throws IllegalArgumentException, if n does not match a valid light
    */
   synchronized native float getLightGreenValue(long hgraph,int n);
   
   /**
    * Returns the blue component of light #n color.
    * @param hgraph handle to the graph native instance
    * @return blue component of color of light #n
    * @throws IllegalArgumentException, if n does not match a valid light
    */
   synchronized native float getLightBlueValue(long hgraph,int n);

   /**
    * Returns the X range low boundary.
    * @param hgraph handle to the graph native instance
    * @return low boundary on X axis
    */
   synchronized native double getRangeX0(long hgraph);

   /**
    * Returns the X range high boundary.
    * @param hgraph handle to the graph native instance
    * @return high boundary on X axis
    */
   synchronized native double getRangeX1(long hgraph);

   /**
    * Returns the Y range low boundary.
    * @param hgraph handle to the graph native instance
    * @return low boundary on Y axis
    */
   synchronized native double getRangeY0(long hgraph);

   /**
    * Returns the Y range high boundary.
    * @param hgraph handle to the graph native instance
    * @return high boundary on Y axis
    */
   synchronized native double getRangeY1(long hgraph);  

   /**
    * Checks, if this is a f(x,y,z)-style function.
    * @param hgraph handle to the graph native instance
    * @return true, if this graph allows z value setting
    */
   synchronized native boolean hasZ(long hgraph);
   
   /**
    * Checks mode bits.
    * @param hgraph handle to the graph native instance
    * @param m the mode bit(s) to check for.
    * @return the mode bits matched
    */
   synchronized native int mode(long hgraph,int m);
   
   /**
    * Rotates the scene around axis given by cross product
    * of reverse projected vectors (x0,y0,0) and (x1,y1,0).
    * @param hgraph handle to the graph native instance
    * @param c the canvas this graph belongs to
    * @param x0 x component of first vector
    * @param y0 y component of first vector
    * @param x1 x component of second vector
    * @param y1 y component of second vector
    */
   synchronized native void screenRotate(long hgraph,Canvas c,int x0,int y0,int x1,int y1);

   /**
    * Sets the background color.
    * @param hgraph handle to the graph native instance
    * @param c the canvas this graph belongs to
    * @param r red value of color
    * @param g green value of color
    * @param b blue value of color
    */
   synchronized native void setBackgroundColor(long hgraph,Canvas c,float r,float g,float b);

   /**
    * Sets the plot range on the X axis.
    * @param hgraph handle to the graph native instance
    * @param c the canvas this graph belongs to
    * @param x0 low bound of range
    * @param x1 high bound of range
    */
   synchronized native void setRangeX(long hgraph,Canvas c,double x0,double x1)
      throws IllegalArgumentException;

   /**
    * Sets the plot range on the Y axis.
    * @param hgraph handle to the graph native instance
    * @param c the canvas this graph belongs to
    * @param y0 low bound of range
    * @param y1 high bound of range
    */
   synchronized native void setRangeY(long hgraph,Canvas c,double y0,double y1)
      throws IllegalArgumentException;

   /**
    * Set the Z parameter for a f(x,y,z)-type function.
    * @param hgraph handle to the graph native instance
    * This may be used for animation purposes.
    * @param z the Z parameter for f(x,y,z)
    */
   synchronized native void setZ(long hgraph,double z);

   /**
    * Sets the r,g,b color of light #n.
    * @param hgraph handle to the graph native instance
    * @param c the canvas this graph belongs to
    * @param n the light number
    * @param r red value of color
    * @param g green value of color
    * @param b blue value of color
    * @throws IllegalArgumentException, if n is out of bounds.
    */
   synchronized native void setLightColor(long hgraph,Canvas c,int n,float r,float g,float b)
      throws IllegalArgumentException;

   /**
    * Sets the default rendering mode (POINTS).
    * @param hgraph handle to the graph native instance
    * @param c the canvas this graph belongs to
    */
   synchronized native void setDefaultMode(long hgraph,Canvas c);

   /**
    * Sets the rendering mode.
    * @param hgraph handle to the graph native instance
    * @param c the canvas this graph belongs to
    * @param mode POINTS, LINES, SMOOTH, LIGHTS, AXES 
    */
   synchronized native void setMode(long hgraph,Canvas c,int mode);

   /**
    * Rotates the graph around x axis.
    * @param hgraph handle to the graph native instance
    * @param c the canvas this graph belongs to
    * @param ax the angle to rotate
    */
   synchronized native void rotateX(long hgraph,Canvas c,double ax);

   /**
    * Rotates the graph around y axis.
    * @param hgraph handle to the graph native instance
    * @param c the canvas this graph belongs to
    * @param ay the angle to rotate
    */
   synchronized native void rotateY(long hgraph,Canvas c,double ay);

   /**
    * Rotates the graph around z axis.
    * @param hgraph handle to the graph native instance
    * @param c the canvas this graph belongs to
    * @param az the angle to rotate
    */
   synchronized native void rotateZ(long hgraph,Canvas c,double az);
   
   /**
    * Unsets the rendering mode bit mode.
    * @param hgraph handle to the graph native instance
    * @param c the canvas this graph belongs to
    * @param mode POINTS, LINES, SMOOTH, LIGHTS, AXES 
    */
   synchronized native void unsetMode(long hgraph,Canvas c,int mode);

   
   /**
    * Zooms the graph in or out.
    * @param hgraph handle to the graph native instance
    * @param c the canvas this graph belongs to
    * @param tz zoom - values less than 0 will zoom out
    *                  values greater than 0 will zoom in
    */
   synchronized native void zoom(long hgraph,Canvas c,double tz);

   /**
    * Frees the native instance.
    * @param hgraph handle to the graph native instance
    * @param c the canvas this graph belongs to
    */
   synchronized native void dispose(long hgraph,Canvas c);

   /**
    * Begin a new jgraph vm function definition.
    * @param name the function name
    * @param signature the function signature
    * @param argc the number of arguments for the function
    * @return data buffer handle of this function
    */
   synchronized native int beginFunction(String name,String signature,int argc);

   /**
    * Adds an instruction to the current function definition.
    * @param ins the instruction
    */
   synchronized native int addIns(int ins);
   
   /**
    * Adds an instruction to the current function definition.
    * @param ins the instruction
    * @param arg the instruction argument
    */
   synchronized native int addIns(int ins,int arg);
   
   /**
    * Sets a label for a branch instruction at address addr.
    * @param addr the branch instruction address
    * @param label the label to branch to
    */
   synchronized native void setLabel(int addr,int label);

   /**
    * Ends the definition of the current function.
    */   
   synchronized native void endFunction()
      throws DataBufferExceeded;

   /**
    * Resets the code generator.
    */   
   synchronized native void reset();

   /**
    * Returns the current function.
    * @return a handle to the native instance
    */   
   synchronized native long getFunction();
   
   /**
    * Registers a double value as a constant.
    * @return data buffer handle of constant
    */
   synchronized native int constant(double c)
      throws DataBufferExceeded;
   
   /**
    * Registers a long value as a constant.
    * @return data buffer handle of constant
    */
   synchronized native int constant(long c)
      throws DataBufferExceeded;
   
   /**
    * Returns the name of the function passed as handle.
    * @param hfunction handle to the function native instance
    * @return the name of the function
    */
   synchronized native String getFunctionName(long hfunction);
   
   /**
    * Returns a function's internal type signature.
    * @param hfunction handle to the function native instance
    * @return the signature of the function
    */
   synchronized native String getFunctionType(long hfunction);
   
   /**
    * Disposes a native function object.
    */
   synchronized native void disposeFunction(long hfunction);

}
