package graph3d.peer;

// (C) by Thalion Graphic Solutions e.K. 2001
//
// Permissions to copy, modify and redistribute
// granted, according to the conditions defined
// in the Gnu Public License (GPL).
// Please read the file gpl.txt provided with this
// distribution.

import graph3d.Function;
import graph3d.RuntimeException;
import graph3d.TypeMismatchException;
import java.awt.Canvas;

/**
 * Peer class for Graph3d.
 * It is the job of 
 */
public class Graph3d {

   /** 
    * The context of this Graph3d.
    */
   private static Context context_;

   /**
    * The native object handle.
    */
   private long handle_;

   static {
      context_ = Context.getInstance();
   }

   /**
    * Mode bit: draw graph as point set.
    */
   public static final int POINTS   = 1;
 
   /**
    * Mode bit: draw graph as wireframe model.
    */
   public static final int LINES    = 2;

   /**
    * Mode bit: draw graph smooth shaded.
    */
   public static final int SMOOTH   = 4;

   /**
    * Mode bit: use lighting.
    */
   public static final int LIGHTS   = 8;

   /**
    * Mode bit: show x,y,z axes.
    */
   public static final int AXES     = 16;
   
   /**
    * Mode bit: use display lists.
    */
   public static final int DLISTS = 32;

   /**
    * Mode bit: draw contour colors.
    */
   public static final int CONTOURS = 64;

   /**
    * Create an instance.
    * @param f the function.
    * @param x0 low bound of x range
    * @param x1 high bound of x range
    * @param y0 low bound of y range
    * @param y1 high bound of y range
    * @param mode the mode bits to use  
    */
   public Graph3d(Function f,
                  double x0,
                  double x1,
                  double y0,
                  double y1,
                  int mode)
      throws IllegalArgumentException,
             RuntimeException,
             TypeMismatchException
   {
      handle_ = context_.create(f,x0,x1,y0,y1,mode);
   }

   /**
    * Create an instance using default bounds of x:-1 to 1 and y:-1 to 1.
    * @param f the function.
    */
   public Graph3d(Function f)
      throws IllegalArgumentException,
             RuntimeException,
             TypeMismatchException
   {
      this(f,-1d,1d,-1d,1d,POINTS|DLISTS);      
   }
   
   /**
    * This method initializes the graph's color parameters to non-default
    * values. It should be invoked before the graph is realized.
    * Use the setBackgroundColor() and setLightColor() methods to alter
    * color parameters later.
    * @param l0r light #0 red component
    * @param l0g light #0 green component
    * @param l0b light #0 blue component
    * @param l1r light #1 red component
    * @param l1g light #1 green component
    * @param l1b light #1 blue component
    * @param bgr background/clear color red component
    * @param bgg background/clear color green component
    * @param bgb background/clear color blue component
    */
   public void initColors(float l0r,float l0g,float l0b,
                          float l1r,float l1g,float l1b,
                          float bgr,float bgg,float bgb)
   {
      context_.initColors(handle_,l0r,l0g,l0b,l1r,l1g,l1b,bgr,bgg,bgb);
   }

   /**
    * This method realizes a graph instance, ie. it actually creates
    * graphics data structures from the function data grid.
    * @param c the canvas this graph belongs to
    */
   public void realize(Canvas c)
   {
      context_.realize(handle_,c);
   }

   /**
    * Display the graph.
    * This method will (and should) be invoked from
    * within a canvas' paint() method. It may not be called
    * from outside the AWT event dispatch thread.
    * @param c the canvas to draw into
    */
   public void display(Canvas c)
   {
      context_.display(handle_,c);
   }

   /**
    * Reshape the graph.
    * This method should be invoked, whenever a viewport resize
    * has occurred.
    * @param c the canvas which owns this graph
    * @param w the new viewport width in screen coordinates
    * @param h the new viewport height in screen coordinates
    */
   public void reshape(Canvas c,int w,int h)
   {
      context_.reshape(handle_,c,w,h);
   }

   /**
    * Support for printing: create a offscreen buffer
    * containing image data for one frame.
    * @param c the canvas this graph belongs to
    * @param w width of printable area
    * @param h height of printable area
    * @return an array of ints representing the buffer
    */
   public int[] getImageData(Canvas c,int w,int h)
   {
      return context_.getImageData(handle_,c,w,h);
   }
   
   /**
    * Returns the red component of light #n color.
    * @return red component of color of light #n
    * @throws IllegalArgumentException, if n does not match a valid light
    */
   public float getLightRedValue(int n)
   {
      return context_.getLightRedValue(handle_,n);
   }
   
   /**
    * Returns the green component of light #n color.
    * @return green component of color of light #n
    * @throws IllegalArgumentException, if n does not match a valid light
    */
   public float getLightGreenValue(int n)
   {
      return context_.getLightGreenValue(handle_,n);
   }
   
   /**
    * Returns the blue component of light #n color.
    * @return blue component of color of light #n
    * @throws IllegalArgumentException, if n does not match a valid light
    */
   public float getLightBlueValue(int n)
   {
      return context_.getLightBlueValue(handle_,n);
   }
   
   /**
    * Returns the X range low boundary.
    * @return low boundary on X axis
    */
   public double getRangeX0()
   {
      return context_.getRangeX0(handle_);
   }

   /**
    * Returns the X range high boundary.
    * @return high boundary on X axis
    */
   public double getRangeX1()
   {
      return context_.getRangeX1(handle_);
   }   

   /**
    * Returns the Y range low boundary.
    * @return low boundary on Y axis
    */
   public double getRangeY0()
   {
      return context_.getRangeY0(handle_);
   }

   /**
    * Returns the Y range high boundary.
    * @return high boundary on Y axis
    */
   public double getRangeY1()
   {
      return context_.getRangeY1(handle_);
   } 
   
   /**
    * Checks, if this is a f(x,y,z)-style function.
    * @return true, if this graph allows z value setting
    */
   public boolean hasZ()
   {
      return context_.hasZ(handle_);
   }
   
   
   /**
    * Checks mode bits.
    * @param m the mode bit(s) to check for.
    * @return the mode bits matched
    */
   public int mode(int m)
   {
      return context_.mode(handle_,m);
   }
   
   /**
    * Rotates the scene around axis given by cross product
    * of reverse projected vectors (x0,y0,0) and (x1,y1,0).
    * @param c the canvas this graph belongs to
    * @param x0 x component of first vector
    * @param y0 y component of first vector
    * @param x1 x component of second vector
    * @param y1 y component of second vector
    */
   public void screenRotate(Canvas c,int x0,int y0,int x1,int y1)
   {
      context_.screenRotate(handle_,c,x0,y0,x1,y1);
   }

   /**
    * Sets the plot range on the X axis.
    * @param c the canvas this graph belongs to
    * @param x0 low bound of range
    * @param x1 high bound of range
    * @throws IllegalArgumentException, if x0 >= x1
    */
   public void setRangeX(Canvas c,double x0,double x1)
      throws IllegalArgumentException
   {
      context_.setRangeX(handle_,c,x0,x1);
   }

   /**
    * Sets the plot range on the Y axis.
    * @param c the canvas this graph belongs to
    * @param y0 low bound of range
    * @param y1 high bound of range
    * @throws IllegalArgumentException, if y0 >= y1
    */
   public void setRangeY(Canvas c,double y0,double y1)
      throws IllegalArgumentException
   {
      context_.setRangeY(handle_,c,y0,y1);   
   }

   /**
    * Set the Z parameter for a f(x,y,z)-type function.
    * This may be used for animation purposes.
    * @param z the Z parameter for f(x,y,z)
    */
   public void setZ(double z)
   {
      context_.setZ(handle_,z);    
   }

   /**
    * Sets the background color.
    * @param c the canvas this graph belongs to
    * @param r red value of color
    * @param g green value of color
    * @param b blue value of color
    * @throws IllegalArgumentException, if the light number is invalid
    */
   public void setBackgroundColor(Canvas c,float r,float g,float b)
   {
      context_.setBackgroundColor(handle_,c,r,g,b);
   }

   /**
    * Sets the color of light #n.
    * @param c the canvas this graph belongs to
    * @param n the light number
    * @param r red value of color
    * @param g green value of color
    * @param b blue value of color
    * @throws IllegalArgumentException, if the light number is invalid
    */
   public void setLightColor(Canvas c,int n, float r,float g,float b)
      throws IllegalArgumentException
   {
      context_.setLightColor(handle_,c,n,r,g,b);
   }

   /**
    * Sets the default rendering mode:
    * no lighting, axes. Point set display is on.
    * @param c the canvas this graph belongs to
    */
   public void setDefaultMode(Canvas c)
   {
      context_.setDefaultMode(handle_,c);
   }

   /**
    * Sets the rendering mode.
    * @param c the canvas this graph belongs to
    * @param mode POINTS, LINES or SMOOTH
    */
   public void setMode(Canvas c,int mode)
   {
      context_.setMode(handle_,c,mode);
   }

   /**
    * Rotates the graph around x axis.
    * @param c the canvas this graph belongs to
    * @param ax the angle to rotate
    */
   public void rotateX(Canvas c,double ax)
   {
      context_.rotateX(handle_,c,ax);   
   }

   /**
    * Rotates the graph around y axis.
    * @param c the canvas this graph belongs to
    * @param ay the angle to rotate
    */
   public void rotateY(Canvas c,double ay)
   {
      context_.rotateY(handle_,c,ay);      
   }

   /**
    * Rotates the graph around z axis.
    * @param c the canvas this graph belongs to
    * @param az the angle to rotate
    */
   public void rotateZ(Canvas c,double az)
   {
      context_.rotateZ(handle_,c,az);      
   }
   
   /**
    * Unsets the rendering mode bit mode.
    * @param c the canvas this graph belongs to
    * @param mode POINTS, LINES or SMOOTH
    */
   public void unsetMode(Canvas c,int mode)
   {
      context_.unsetMode(handle_,c,mode);
   }
   
   /**
    * Zooms the graph.
    * @param tz - positive values mean zoom out
    *             negative zoom in
    */
   public void zoom(Canvas c,double tz)
   {
      context_.zoom(handle_,c,tz);      
   }

   /**
    * Frees the native instance.
    * @param c the canvas this graph belongs to
    */
   public void dispose(Canvas c)
   {
       context_.dispose(handle_,c);
   }

}
