package ui;

import java.awt.Color;
import java.awt.Component;
import java.awt.EventQueue;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.Box;
import javax.swing.ButtonGroup;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JColorChooser;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JSeparator;
import javax.swing.JTextField;
import javax.swing.JToolBar;

import javax.swing.colorchooser.*;
import javax.swing.event.ChangeListener;

import java.util.Enumeration;
import java.util.Vector;

import graph3d.GraphView;

/**
 * This class implements a document-specific
 * tools panel.
 * The panel contains buttons to select view state flags,
 * and input fields to specify plot ranges.
 */
public class DocumentTools {

   private static class LightColorSelectionModel
      implements ColorSelectionModel {
      private int n_;
      public LightColorSelectionModel(int n)
      {
         n_ = n;
      }
      public Color getSelectedColor()
      {
         if ( DocumentWindow.getCurrentDocument() != null ) {
	    int r = (int) (255 * DocumentWindow.getCurrentDocument().getView().getLightRedValue(n_)),
	        g = (int) (255 * DocumentWindow.getCurrentDocument().getView().getLightGreenValue(n_)),
	        b = (int) (255 * DocumentWindow.getCurrentDocument().getView().getLightBlueValue(n_));
            return new Color(r,g,b);
	 }
	 return Color.white;
      }
      public void setSelectedColor(Color color)
      {
         if ( DocumentWindow.getCurrentDocument() != null ) {
	    DocumentWindow.getCurrentDocument().getView().setLightColor(n_,color.getRed()/255f,color.getGreen()/255f,color.getBlue()/255f);
	 }
      }
      public void addChangeListener(ChangeListener listener)
      {
         // Does not care about listeners.
      }
      public void removeChangeListener(ChangeListener listener)
      {
         // Does not care about listeners.
      }       
   }
   
   private static JDialog[]             choosers_;
   private static ColorSelectionModel[] models_;
    
   static {
      models_ = new LightColorSelectionModel[2];
      models_[0] = new LightColorSelectionModel(0);
      models_[1] = new LightColorSelectionModel(1);
      choosers_ = new JDialog[2];
      ActionListener deactivator0 = new ActionListener() {
	 public void actionPerformed(ActionEvent e)
         {
            DocumentWindow.getCurrentDocument().getTools().setLightColorChooserSelected(0,false); 
         }
      };
      ActionListener deactivator1 = new ActionListener() {
	 public void actionPerformed(ActionEvent e)
         {
            DocumentWindow.getCurrentDocument().getTools().setLightColorChooserSelected(1,false); 
         }
      };
      choosers_[0] = JColorChooser.createDialog(App.getInstance(),
                                                ResourceManager.getDialogString("set-light0"),
			                        false,
				                new JColorChooser(models_[0]),
				                deactivator0,
				                deactivator0);
      choosers_[1] = JColorChooser.createDialog(App.getInstance(),
                                                ResourceManager.getDialogString("set-light1"),
			                        false,
				                new JColorChooser(models_[1]),
				                deactivator1,
				                deactivator1);
   }    

   private DocumentWindow doc_;
   private GraphView      view_;
   private double         x0_, x1_, y0_, y1_, z0_, z1_;
   private DecimalField     x0_input_,
                            x1_input_,
		            y0_input_,
		            y1_input_,
		            z0_input_,
		            z1_input_;
		      
   private Vector         controls_;

   private JCheckBox[]    light_cb_;

   /**
    * Creates toolbar control set for the document.
    * The controls created depend on the document
    * type. Ternary function control panels will contain
    * additional components for z value and animation control.
    * @param document the document view
    * @return a tool panel.
    */
   public DocumentTools(DocumentWindow document)
   {
      doc_ = document;
      view_ = document.getView();
      controls_ = new Vector();
      light_cb_ = new JCheckBox[2];
      x0_ = view_.getRangeX0();
      x1_ = view_.getRangeX1();
      y0_ = view_.getRangeY0();
      y1_ = view_.getRangeY1();
      addCommonControls();
      if ( view_.hasZ() ) {
         z0_ = Preferences.getInstance().getGraphRangeZ0();
         z1_ = Preferences.getInstance().getGraphRangeZ1();
         addZControls();
      }
   }
   
   /**
    * Writes the control set to a given toolbar.
    * Previously contained controls will be removed
    * from the toolbar and replaced by the document tools.
    * @param toolbar the toolbar
    */
   public void setToolBar(JToolBar toolbar)
   {
      Enumeration e = controls_.elements();
      while ( e.hasMoreElements() ) {
         toolbar.add((Component)e.nextElement());
      }
   }
   
   private void addCommonControls()
   {
      JPanel panel = new JPanel();
      JButton b = new JButton();
      ResourceManager.setIconForButton("printer",b);
      b.setToolTipText(ResourceManager.getToolTip("print"));
      b.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent event) {
            doc_.print();
         }
      });
      panel.add(b);
      JCheckBox cb = new JCheckBox();
      ResourceManager.setIconForButton("axes",cb);
      cb.setToolTipText(ResourceManager.getToolTip("draw-axes"));
      cb.setSelected(view_.mode(GraphView.AXES) != 0);
      cb.addItemListener(new ItemListener() {
         public void itemStateChanged(ItemEvent event) {
	    if ( event.getStateChange() == ItemEvent.SELECTED ) {
               view_.setMode(GraphView.AXES);
	    }
	    else {
	       view_.unsetMode(GraphView.AXES);
	    }
         }
      });
      panel.add(cb);
      cb = new JCheckBox();
      ResourceManager.setIconForButton("contours",cb);
      cb.setToolTipText(ResourceManager.getToolTip("use-contours"));
      cb.setSelected(view_.mode(GraphView.CONTOURS) != 0);	
      cb.addItemListener(new ItemListener() {
         public void itemStateChanged(ItemEvent event) {
	    if ( event.getStateChange() == ItemEvent.SELECTED ) {
               view_.setMode(GraphView.CONTOURS);
	    }
	    else {
	       view_.unsetMode(GraphView.CONTOURS);
	    }
         }
      });
      panel.add(cb);
      add(panel);
      add(new JSeparator());

      panel = new JPanel();
      cb = new JCheckBox();
      ResourceManager.setIconForButton("lighting",cb);
      cb.setToolTipText(ResourceManager.getToolTip("use-lighting"));
      cb.setSelected(view_.mode(GraphView.LIGHTS) != 0);	
      cb.addItemListener(new ItemListener() {
         public void itemStateChanged(ItemEvent event) {
	    if ( event.getStateChange() == ItemEvent.SELECTED ) {
               view_.setMode(GraphView.LIGHTS);
	    }
	    else {
	       view_.unsetMode(GraphView.LIGHTS);
	    }
         }
      });
      panel.add(cb);
      
      light_cb_[0] = cb = new JCheckBox();
      ResourceManager.setIconForButton("light0",cb);
      cb.setToolTipText(ResourceManager.getToolTip("edit-light0"));
      cb.setSelected(showingColorChooser(0));	
      cb.addItemListener(new ItemListener() {
         public void itemStateChanged(ItemEvent event) {
	    if ( event.getStateChange() == ItemEvent.SELECTED ) {
	       showColorChooser(0);
	    }
	    else {
	       hideColorChooser(0);
	    }
         }
      });
      panel.add(cb);

      light_cb_[1] = cb = new JCheckBox();
      ResourceManager.setIconForButton("light1",cb);
      cb.setToolTipText(ResourceManager.getToolTip("edit-light1"));
      cb.setSelected(showingColorChooser(1));	
      cb.addItemListener(new ItemListener() {
         public void itemStateChanged(ItemEvent event) {
	    if ( event.getStateChange() == ItemEvent.SELECTED ) {
	       showColorChooser(1);
	    }
	    else {
	       hideColorChooser(1);
            }
         }
      });
      panel.add(cb);
      add(panel);
      add(new JSeparator());
	
      panel = new JPanel();
      ButtonGroup group = new ButtonGroup();
		
      JRadioButton rb = new JRadioButton();
      ResourceManager.setIconForButton("points",rb);
      rb.setToolTipText(ResourceManager.getToolTip("draw-points"));
      rb.setSelected(view_.mode(GraphView.POINTS) != 0);	
      rb.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            view_.setMode(GraphView.POINTS);
	    view_.unsetMode(GraphView.LINES);
	    view_.unsetMode(GraphView.SMOOTH);
         }
      });
      panel.add(rb);
      group.add(rb);
	
      rb = new JRadioButton();
      ResourceManager.setIconForButton("lines",rb);
      rb.setToolTipText(ResourceManager.getToolTip("draw-lines"));
      rb.setSelected(view_.mode(GraphView.LINES) != 0);	
      rb.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            view_.setMode(GraphView.LINES);
	    view_.unsetMode(GraphView.SMOOTH);
         }
      });
      panel.add(rb);
      group.add(rb);
	
      rb = new JRadioButton();
      ResourceManager.setIconForButton("smooth",rb);
      rb.setToolTipText(ResourceManager.getToolTip("draw-smooth"));
      rb.setSelected(view_.mode(GraphView.SMOOTH) != 0);	
      rb.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            view_.setMode(GraphView.SMOOTH);
         }
      });
      panel.add(rb);
      group.add(rb);
      add(panel);
      add(new JSeparator());
	
      JLabel label = new JLabel("X:");
      add(label);
      // Sorry, dear Swing authors. I don't have years of time at my
      // hand to figure out how to prevent your damn text fields of being
      // margin-resized to idiotic dimensions, without regard of adding
      // glue objects to the toolbar or custom insets to the text fields.
      // So each of them gets a panel of it's own.
      panel = new JPanel();
      x0_input_ = new DecimalField(x0_,10);
      x0_input_.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent event )
	 {
	    double x0 = x0_input_.getValue();
	    try {
               view_.setRangeX(x0,x1_);
	       x0_ = x0;
            }
	    catch ( IllegalArgumentException e ) {
	       JOptionPane.showMessageDialog(null,
		                             "x0 > x1",
					     ResourceManager.getErrorMessage("invalid-range"),
					     JOptionPane.ERROR_MESSAGE);	     
	    }
	 }
      });
      panel.add(x0_input_);
      add(panel);
      label = new JLabel("-");
      add(label);
      panel = new JPanel();
      x1_input_ = new DecimalField(x1_,10);
      x1_input_.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent event )
	 {
	    double x1 = x1_input_.getValue();
	    try {
               view_.setRangeX(x0_,x1);
	       x1_ = x1;
            }
	    catch ( IllegalArgumentException e ) {
	       JOptionPane.showMessageDialog(null,
		                             "x1 < x0",
					     ResourceManager.getErrorMessage("invalid-range"),
					     JOptionPane.ERROR_MESSAGE);	     
	    }
	 }
      });
      panel.add(x1_input_);
      add(panel);
      label = new JLabel("Y:");
      add(label);
      panel = new JPanel();
      y0_input_ = new DecimalField(y0_,10);
      y0_input_.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent event )
	 {
	    double y0 = y0_input_.getValue();
	    try {
               view_.setRangeY(y0,y1_);
	       y0_ = y0;
            }
	    catch ( IllegalArgumentException e ) {
	       JOptionPane.showMessageDialog(null,
		                             "y0 > y1",
					     ResourceManager.getErrorMessage("invalid-range"),
					     JOptionPane.ERROR_MESSAGE);	     
	    }
	 }       
      });
      panel.add(y0_input_);
      add(panel);
      label = new JLabel("-");
      add(label);
      panel = new JPanel();
      y1_input_ = new DecimalField(y1_,10);
      y1_input_.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent event)
	 {
	    double y1 = y1_input_.getValue();
	    try {
               view_.setRangeY(y0_,y1);
	       y1_ = y1;
            }
	    catch ( IllegalArgumentException e ) {
	       JOptionPane.showMessageDialog(null,
		                             "y1 < y0",
					     ResourceManager.getErrorMessage("invalid-range"),
					     JOptionPane.ERROR_MESSAGE);	     
	    }
	 }       
      });
      panel.add(y1_input_);
      add(panel);
      add(Box.createVerticalGlue());
   }
   
   private void addZControls()
   {
      add(new JSeparator());
      JLabel label = new JLabel("Z:");
      add(label);
      JPanel panel = new JPanel();
      z0_input_ = new DecimalField(z0_,10);
      z0_input_.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent event)
	 {
	    double z0 = z0_input_.getValue();
            if ( z0 < z1_ ) {
	       z0_ = z0;
	    }
	    else {
	       JOptionPane.showMessageDialog(null,
		                             "z0 > z1",
					     ResourceManager.getErrorMessage("invalid-range"),
					     JOptionPane.ERROR_MESSAGE);		
            }
	 }
      });
      panel.add(z0_input_);
      add(panel);
      label = new JLabel("-");
      add(label);
      panel = new JPanel();
      z1_input_ = new DecimalField(z1_,10);
      z1_input_.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent event)
	 {
	    double z1 = z1_input_.getValue();
            if ( z1 > z0_ ) {
	       z1_ = z1;
            }
	    else {
	       JOptionPane.showMessageDialog(null,
		                             "z1 < z0",
					     ResourceManager.getErrorMessage("invalid-range"),
					     JOptionPane.ERROR_MESSAGE);
	    }
	 }
      });
      panel.add(z1_input_);
      add(panel);
      add(new JSeparator());
	
      panel = new JPanel();	
      ButtonGroup group = new ButtonGroup();
		
      JRadioButton rb = new JRadioButton();
      ResourceManager.setIconForButton("none",rb);
      rb.setToolTipText(ResourceManager.getToolTip("animate-none"));
      rb.setSelected(true);	
      rb.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            doc_.stopAnimation();
         }
      });
      panel.add(rb);
      group.add(rb);
	
      rb = new JRadioButton();
      ResourceManager.setIconForButton("loop",rb);
      rb.setToolTipText(ResourceManager.getToolTip("animate-loop"));
      rb.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            doc_.loopAnimate(z0_,z1_,(z1_ - z0_) / 100);
         }
      });
      panel.add(rb);
      group.add(rb);
	
      rb = new JRadioButton();
      ResourceManager.setIconForButton("swing",rb);
      rb.setToolTipText(ResourceManager.getToolTip("animate-swing"));	
      rb.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            doc_.swingAnimate(z0_,z1_,(z1_ - z0_) / 100);
         }
      });
      panel.add(rb);
      group.add(rb);
      add(panel);                  
   }
   
   private void add(Component c)
   {
      controls_.addElement(c);
   }

   public void setLightColorChooserSelected(int n,boolean flag)
   {
      light_cb_[n].setSelected(flag);
   }
   
   public void showColorChooser(int n)
   {
      choosers_[n].show();
   }
   
   public void hideColorChooser(int n)
   {
      choosers_[n].hide();
   }
   
   public boolean showingColorChooser(int n)
   {
      return choosers_[n].isVisible();
   }

}
