package ui;


import javax.swing.JDialog;
import javax.swing.JInternalFrame;
import javax.swing.JOptionPane;
import javax.swing.JToolBar;

import javax.swing.event.ChangeListener;
import javax.swing.event.InternalFrameListener;
import javax.swing.event.InternalFrameEvent;

import java.awt.*;
import java.awt.PrintJob;
import java.awt.Toolkit;

import java.awt.print.PageFormat;
import java.awt.print.PrinterJob;

import graph3d.Function;
import graph3d.GraphView;
import graph3d.RuntimeException;
import graph3d.TypeMismatchException;

public class DocumentWindow
    extends JInternalFrame
    implements InternalFrameListener {

   /**
    * Animation delay in ms.
    */
   public static final int DELAY  = 250;

   private class LoopAnimator
      extends Thread {
      
      double z0_, z1_, step_;
   
      public LoopAnimator(double z0,double z1,double step)
      {
         super("Loop Animator");
	 setPriority(MIN_PRIORITY);
	 z0_ = z0;
	 z1_ = z1;
	 step_ = step;
      }
      
      public void run()
      {
         double z = z0_;
         for ( ;; ) {
	    try {
	       view_.setZ(z);
	       Runnable r = new Runnable() {
	          public void run()
		  {
	             view_.paint(null);		  
		  }
	       };
               EventQueue.invokeLater(r);	       
	       z += step_;
	       if ( z > z1_ ) {
	          z = z0_;
	       }
	       Thread.sleep(DELAY);
	    }
	    catch ( InterruptedException e ) {}
	 }
      }
      
   }
   
   private class SwingAnimator
      extends Thread {
      
      double z0_, z1_, step_;
   
      public SwingAnimator(double z0,double z1,double step)
      {
         super("Swing Animator");
	 setPriority(MIN_PRIORITY);
	 z0_ = z0;
	 z1_ = z1;
	 step_ = step;
      }
      
      public void run()
      {
         double z = z0_;
         for ( ;; ) {
	    try {
	       view_.setZ(z);
	       Runnable r = new Runnable() {
	          public void run()
		  {
	             view_.paint(null);		  
		  }
	       };
               EventQueue.invokeLater(r);
	       z += step_;
	       if ( z > z1_ || z < z0_ ) {
	          step_ = -step_;
	       }
	       Thread.sleep(DELAY);
	    }
	    catch ( InterruptedException e ) {}
	 }
      }
      
   }

    private static DocumentWindow currentDocument_;

    static int openFrameCount = 0;
    static final int xOffset = 30, yOffset = 30;

    private DocumentTools  tools_;    
    private GraphView      view_;
    private Thread         animator_;
    
    public DocumentWindow(Function f)
       throws IllegalArgumentException,
              RuntimeException,
	      TypeMismatchException     
    {
        super(f.toString(), 
              true, //resizable
              true, //closable
              true, //maximizable
              true);//iconifiable

        //...Create the GUI and put it in the window...
        getContentPane().setLayout(new BorderLayout());
        view_ = new GraphView(f,
                              Preferences.getInstance().getGraphRangeX0(),
                              Preferences.getInstance().getGraphRangeX1(),
                              Preferences.getInstance().getGraphRangeY0(),
                              Preferences.getInstance().getGraphRangeY1(),
                              Preferences.getInstance().getGraphModeBits());
        view_.initColors(Preferences.getInstance().getGraphLightColor(0),
                         Preferences.getInstance().getGraphLightColor(1),
                         Preferences.getInstance().getGraphBackgroundColor());
        getContentPane().add(view_);

        //...Then set the window size or call pack...
        setSize(300,300);

        //Set the window's location.
	openFrameCount ++;
        setLocation(xOffset*openFrameCount, yOffset*openFrameCount);
	addInternalFrameListener(this);
	tools_ = new DocumentTools(this);
	currentDocument_ = this;
    }
    
    public GraphView getView()
    {
       return view_;
    }
    
    public DocumentTools getTools()
    {
       return tools_;
    }
    
    public static DocumentWindow getCurrentDocument()
    {
       return currentDocument_;
    }

    private static void setCurrentDocument(DocumentWindow docwin)
    {
       if ( currentDocument_ != null ) {
          currentDocument_.getTools().hideColorChooser(0);
          currentDocument_.getTools().hideColorChooser(1);
       }
       currentDocument_ = docwin;
    }
    
    public void makeCurrent()
    {

       setCurrentDocument(this);
       App.getInstance().setTools(getTools());
    }
    
    public void loopAnimate(double z0,double z1,double step)
    {
       if ( animator_ != null ) {
          animator_.stop();
       }
       animator_ = new LoopAnimator(z0,z1,step);
       animator_.start();
    }
    
    public void swingAnimate(double z0,double z1,double step)
    {
       if ( animator_ != null ) {
          animator_.stop();
       }
       animator_ = new SwingAnimator(z0,z1,step);
       animator_.start();
    }
    
    public void stopAnimation()
    {
       if ( animator_ != null ) {
          animator_.stop();
	  animator_ = null;
       }
    }
    
    public void internalFrameActivated(InternalFrameEvent e)
    {
        makeCurrent();
    }
    
    public void internalFrameClosed(InternalFrameEvent e)
    {
	stopAnimation();   
        App.getInstance().setTools(null);
        setCurrentDocument(null);     
    }
    
    public void internalFrameClosing(InternalFrameEvent e)
    {
        // Currently unused.
    }
    
    public void internalFrameDeactivated(InternalFrameEvent e)
    {
        setCurrentDocument(null);    
    }
    
    public void internalFrameDeiconified(InternalFrameEvent e)
    {
        makeCurrent();
	if ( animator_ != null ) {
	   animator_.start();
	}      
    }
    
    public void internalFrameIconified(InternalFrameEvent e)
    {
	if ( animator_ != null ) {
	   animator_.stop();
	}   
    }
    
    public void internalFrameOpened(InternalFrameEvent e)
    {
        makeCurrent();    
    }

    public void print()
    {
        try {
            PrinterJob job = PrinterJob.getPrinterJob();
            PageFormat page = job.defaultPage();
            page.setOrientation(PageFormat.LANDSCAPE);
            job.setPrintable(view_,job.pageDialog(page));
            job.setJobName(getTitle());
            if ( job.printDialog() ) {
                System.out.println("Now printing ...");
                job.print();
                System.out.println("Printing done.");
            }
        }
        catch ( Exception exception ) {
            JOptionPane.showMessageDialog(this,
                                          exception.getMessage(),
                                          "Printing failed.",
                                          JOptionPane.ERROR_MESSAGE);
        }
    }

}

