// (C) by Thalion Graphic Solutions e.K. 2001
//
// Permissions to copy, modify and redistribute
// granted, according to the conditions defined
// in the Gnu Public License (GPL).
// Please read the file gpl.txt provided with this
// distribution.

#include <graph/graph3d.h>

#include "plaf.h"
#include "trace.h"

#include <GL/glu.h>
#include <stdlib.h>
#include <math.h>

#ifdef _DEBUG
#include <iostream.h>
#endif

//////////////////////////////////////////////////////////////////////////////////
//
// These members of graph are non-public and
// thus declared here.
//
namespace graph {

   //
   // Helper function: calculate the
   // cross product on 2 vectors.
   //
   void crossprod(double *,double *,double *);   

   //
   // Helper function: calculate the normalized
   // cross product on 2 vectors.
   //
   void normal(double *,double *,double *);

}

//////////////////////////////////////////////////////////////////////////////////
//
// Graphable* member definitions.
//

template <int GW,int GH>
graph::Graph3d<GW,GH> *graph::GraphableFunction2<GW,GH>::getGraph(double x0,
                                                                  double x1,
                                                                  double y0,
                                                                  double y1,
                                                                  int mode)
{
   return new Graph3d<GW,GH> (*this,x0,x1,y0,y0,mode);
}

template <int GW,int GH>
graph::Graph3d<GW,GH> *graph::GraphableFunction3<GW,GH>::getGraph(double x0,
                                                                  double x1,
                                                                  double y0,
                                                                  double y1,
                                                                  int mode)
{
   return new Graph3d<GW,GH> (*this,x0,x1,y0,y0,mode);
}


template <int GW,int GH>
graph::GraphableFunction3<GW,GH>::~GraphableFunction3()
{
   // empty.
}

template <int GW,int GH>
graph::Function3Wrapper<GW,GH>::Function3Wrapper(graph::GraphableFunction3<GW,GH>& f):
   f_(&f),z_(0)
{
   f_->ref();
}

template <int GW,int GH>
double graph::Function3Wrapper<GW,GH>::operator () (double x,double y)
{
   return (*f_)(x,y,z_);
}

template <int GW,int GH>
void graph::Function3Wrapper<GW,GH>::setZ(double z)
{
   z_ = z;
}

template <int GW,int GH>
graph::Function3Wrapper<GW,GH>::~Function3Wrapper()
{
   f_->unref();
}

double graph::DefaultRGBFunc::operator () (double x,double y,double z)
{
   //trace("graph::DefaultRGBFunc::operator (x,y,z)");
   return 1;
}

graph::DefaultRGBFunc *graph::DefaultRGBFunc::getInstance()
{
   if ( ! instance_ ) {
      instance_ = new DefaultRGBFunc();
      instance_->ref();
   }
   return instance_;
}

graph::DefaultRGBFunc *graph::DefaultRGBFunc::instance_ = NULL;

//////////////////////////////////////////////////////////////////////////////////
//
// Graph3d<GW,GH> member definitions.
//

template <int GW,int GH>
void graph::Graph3d<GW,GH>::initLights()
{
   trace("Graph3d::initLights");
   // Create light rgba arrays.
   light0_ambient_ = new float[4];
   light0_color_ = new float[4];
   light1_ambient_ = new float[4];
   light1_color_ = new float[4];  
   // Pre-initialize them with some default settings.
   light0_ambient_[0] = 0.2;
   light0_ambient_[1] = 0.0;
   light0_ambient_[2] = 0.0;
   light0_color_[0] = 1.0;
   light0_color_[1] = 0.0;
   light0_color_[2] = 0.0;
   light1_ambient_[0] = 0.0;
   light1_ambient_[1] = 0.0;
   light1_ambient_[2] = 0.2;
   light1_color_[0] = 0.0;
   light1_color_[1] = 0.0;
   light1_color_[2] = 1.0;
   light0_ambient_[3] =
      light0_color_[3] =
      light1_ambient_[3] =
      light1_color_[3] = 0.0;
}

inline void graph::crossprod(double *a,double *b,double *n)
{
   n[0] = a[1]*b[2] - a[2]*b[1],
   n[1] = a[2]*b[0] - a[0]*b[2],
   n[2] = a[0]*b[1] - a[1]*b[0];
}

inline void graph::normal(double *a,double *b,double *n)
{
   double n1_tmp = a[1]*b[2] - a[2]*b[1],
          n2_tmp = a[2]*b[0] - a[0]*b[2],
          n3_tmp = a[0]*b[1] - a[1]*b[0],
          n_abs = sqrt(n1_tmp * n1_tmp + n2_tmp * n2_tmp + n3_tmp * n3_tmp);
   if ( n3_tmp > 0 ) { // We want all normals to be in positive z!
      n[0] = n1_tmp / n_abs;
      n[1] = n2_tmp / n_abs;
      n[2] = n3_tmp / n_abs;
   }
   else {
      n[0] = -n1_tmp / n_abs;
      n[1] = -n2_tmp / n_abs;
      n[2] = -n3_tmp / n_abs;
   }
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::initGraphValues()
{
   trace("Graph3d::initGraphValues");
   double sx = x1_ - x0_;
   double dx = sx / GW;
   double sy = y1_ - y0_;
   double dy = sy / GH;
   int i = 0, j = 0;
   min_ = 0;
   max_ = 0;
   for ( ; i<GW; i++ ) {
      for ( j=0; j<GH*3; j+=3 ) {
         int k = i * GH*3 + j;
         double x = x0_ + i * dx;
         double y = y0_ + j * dy / 3;
         double z = (*function_)(x,y);
         double z_s = 2 * z / sx;
         if ( z_s < min_ ) {
            min_ = z_s;
         }
         if ( z_s > max_ ) {
            max_ = z_s;
         }
         vertices_[k]   = 2 * x / sx;
         vertices_[k+1] = 2 * y / sy;
         vertices_[k+2] = z_s;
         //printf("(%f,%f,%f)\n",vertices_[k],vertices_[k+1],vertices_[k+2]);
         colors_[k] = (*rfunc_)(x,y,z);
         colors_[k + 1] = (*gfunc_)(x,y,z);
         colors_[k + 2] = (*bfunc_)(x,y,z);
      }
   }
   double v1[3], v2[3];
   for ( i=0; i<(GW-1)*(GH-1)*3; i+=3 ) {
      v1[0] = vertices_[i] - vertices_[i+3];
      v1[1] = vertices_[i+1] - vertices_[i+4];
      v1[2] = vertices_[i+2] - vertices_[i+5];
      v2[0] = vertices_[i+3] - vertices_[i+GH*3];
      v2[1] = vertices_[i+4] - vertices_[i+GH*3+1];
      v2[2] = vertices_[i+5] - vertices_[i+GH*3+2];
      normal(v1,v2,normals_ + i);
   }
   recompile_ = true;
#ifdef _DEBUG
   cout << "vertices:" << endl;
   for ( i=0; i<GW*GH; i++ ) {
      cout << i << ": (" << vertices_[i * 3] << ',' << vertices_[i * 3 + 1] << ',';
      cout << vertices_[i * 3 + 2] << ')' << endl;
   }
   cout << "normals:" << endl;
   for ( i=0; i<(GW-1)*(GH-1); i++ ) {
      cout << i << ": (" << normals_[i * 3] << ',' << normals_[i * 3 + 1];
      cout << ',' << normals_[i * 3 + 2] << ')' << endl;
   }
#endif
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::initGraphIndices()
{
   trace("Graph3d::initGraphIndices");
   int i=0, j=0;
   for ( ; i<GW-1; i++ ) {
      for ( j=0; j<GH-1; j++ ) {
         indices_[(i * (GH-1) + j) * 4] = (i * GH) + j;
         indices_[(i * (GH-1) + j) * 4 + 1] = (i * GH) + j + 1;
         indices_[(i * (GH-1) + j) * 4 + 2] = (i * GH) + j + GH + 1;
         indices_[(i * (GH-1) + j) * 4 + 3] = (i * GH) + j + GH;         
      }
   }
#ifdef _DEBUG
   cout << "indices:" << endl;
   for ( i=0; i<(GW-1)*(GH-1); i++ ) {
      cout << i << ": (" << indices_[i * 4] << ',' << indices_[i * 4 + 1];
      cout << ',' << indices_[i * 4 + 2] << ',' << indices_[i * 4 + 3] << ')' << endl;
   }
#endif
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::initSceneSettings()
{
   trace("Graph3d::initSceneSettings");
   // Material descriptor arrays.
   GLfloat mat_specular[] = { 1.0, 1.0, 1.0, 1.0 };
   GLfloat mat_shininess[] = { 50.0 };
   // Light default positions.
   GLfloat light0_position[] = { 1.0, 1.0, 1.0, 0.0 };
   GLfloat light1_position[] = { -1.0, -1.0, 1.0, 0.0 };
   // Set clear color.
   glClearColor (bg_r_, bg_g_, bg_b_, 0.0);
   // Set material properties.
   glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
   glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
   // Set light properties for light 0.
   glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambient_);
   glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_color_);
   glLightfv(GL_LIGHT0, GL_SPECULAR, light0_color_);
   glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
   // Enable light 0.
   glEnable(GL_LIGHT0);
   // Set light properties for light 1.
   glLightfv(GL_LIGHT1, GL_AMBIENT, light1_ambient_);
   glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_color_);
   glLightfv(GL_LIGHT1, GL_SPECULAR, light1_color_);
   glLightfv(GL_LIGHT1, GL_POSITION, light1_position);
   glLightf(GL_LIGHT1, GL_CONSTANT_ATTENUATION, 1.5);
   glLightf(GL_LIGHT1, GL_LINEAR_ATTENUATION, 0.5);
   glLightf(GL_LIGHT1, GL_QUADRATIC_ATTENUATION, 0.2);
   // Enable light 1.
   glEnable(GL_LIGHT1);
   glEnable(GL_DEPTH_TEST);
   // Enable GL arrays.
   glEnableClientState (GL_VERTEX_ARRAY);
   glEnableClientState(GL_NORMAL_ARRAY);
   glEnableClientState(GL_COLOR_ARRAY);
   glVertexPointer(3, GL_DOUBLE, 0, GW * GH, vertices_);
   glNormalPointer(GL_DOUBLE, 0, (GW - 1) * (GH - 1), normals_);
   glColorPointer(3, GL_DOUBLE, 0, GW * GH, colors_);
}

//
// Create the graph object using a 2-parameter function.
//
template <int GW,int GH>
graph::Graph3d<GW,GH>::Graph3d(graph::GraphableFunction2<GW,GH>& f,
                               double x0,
                               double x1,
                               double y0,
                               double y1,
                               int mode):
   function_(&f),is_function3_(false),
   rfunc_(NULL),gfunc_(NULL),bfunc_(NULL),
   min_(0),max_(0),x0_(x0),x1_(x1),y0_(y0),y1_(y1),
   bg_r_(0),bg_g_(0),bg_b_(0),   
   rx_(0),ry_(0),rz_(0),tz_(-5),
   recompile_(true),
   lists_created_(false),
   mode_(mode)
{
   trace("Graph3d::Graph3d");
   if ( x0 < x1 && y0 < y1 ) {
      rfunc_ = DefaultRGBFunc::getInstance();
      gfunc_ = DefaultRGBFunc::getInstance();
      bfunc_ = DefaultRGBFunc::getInstance();
      rfunc_->ref();
      gfunc_->ref();
      bfunc_->ref();
      function_->ref();
      initLights();
      initGraphValues();
      initGraphIndices();
   }
   else {
      throw illegal_argument_error();
   }
}

//
// Create the graph object using a 2-parameter function.
//
template <int GW,int GH>
graph::Graph3d<GW,GH>::Graph3d(graph::GraphableFunction3<GW,GH>& f,
                               double x0,
                               double x1,
                               double y0,
                               double y1,
                               int mode):
   function_(NULL),is_function3_(true),
   rfunc_(NULL),gfunc_(NULL),bfunc_(NULL),
   min_(0),max_(0),x0_(x0),x1_(x1),y0_(y0),y1_(y1),
   bg_r_(0),bg_g_(0),bg_b_(0),
   rx_(0),ry_(0),rz_(0),tz_(-5),
   recompile_(true),
   lists_created_(false),
   mode_(mode)
{
   trace("Graph3d::Graph3d");
   if ( x0 < x1 && y0 < y1 ) {
      rfunc_ = DefaultRGBFunc::getInstance();
      gfunc_ = DefaultRGBFunc::getInstance();
      bfunc_ = DefaultRGBFunc::getInstance();
      function_ = new Function3Wrapper<GW,GH> (f);
      rfunc_->ref();
      gfunc_->ref();
      bfunc_->ref();
      function_->ref();
      initLights();
      initGraphValues();
      initGraphIndices();
   }
   else {
      throw illegal_argument_error();   
   }
}


//
// Initialize it.
//
template <int GW,int GH>
void graph::Graph3d<GW,GH>::init()
{
   trace("Graph3d::init");
   initSceneSettings();
   if ( mode_ & DLISTS ) {
      createDisplayList();
   }
   // Initialize the modelview matrix.
   glLoadIdentity();
   glGetFloatv(GL_MODELVIEW_MATRIX,modelview_);
   setMode(mode_);
}

//
// Initialize light colors.
//
template <int GW,int GH>
void graph::Graph3d<GW,GH>::initColors(float l0r,float l0g,float l0b,
                                       float l1r,float l1g,float l1b,
                                       float bgr,float bgg,float bgb)
{
   light0_ambient_[0] = l0r / 5;
   light0_ambient_[1] = l0g / 5;
   light0_ambient_[2] = l0b / 5;
   light0_color_[0] = l0r;
   light0_color_[1] = l0g;
   light0_color_[2] = l0b;
   light1_ambient_[0] = l1r / 5;
   light1_ambient_[1] = l1g / 5;
   light1_ambient_[2] = l1b / 5;
   light1_color_[0] = l1r;
   light1_color_[1] = l1g;
   light1_color_[2] = l1b;
   bg_r_ = bgr;
   bg_g_ = bgg;
   bg_b_ = bgb;
}

//
// Create Display lists for both sphere geometries and some
// additional, but constant stuff.
//
template <int GW,int GH>
void graph::Graph3d<GW,GH>::createDisplayList()
{
   trace("Graph3d::createDisplayList"); 
   if ( ! lists_created_ ) {
      // Display lists for both spheres.
      hgraph_ = glGenLists((GLsizei)1);
      haxes_ = glGenLists((GLsizei)1);
      lists_created_ = true;
   }
   glNewList(hgraph_,GL_COMPILE);
      drawGrid();
   glEndList();
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   glNewList(haxes_,GL_COMPILE);
      drawAxes();
   glEndList();
}

//
// Draw the axes.
//
template <int GW,int GH>
void graph::Graph3d<GW,GH>::drawAxes()
{
   trace("Graph3d::drawAxes");
   printf("drawAxes()\n");
   //
   // Patterns for glyphs x,y,z.
   //
   GLubyte x_bits[] = { 0x00, 0x00, 0xc3, 0x66, 0x66, 0x3c, 0x3c, 0x18, 0x3c, 0x3c, 0x66, 0x66, 0xc3 };
   GLubyte y_bits[] = { 0x00, 0x00, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x3c, 0x3c, 0x66, 0x66, 0xc3 };
   GLubyte z_bits[] = { 0x00, 0x00, 0xff, 0xc0, 0xc0, 0x60, 0x30, 0x7e, 0x0c, 0x06, 0x03, 0x03, 0xff };
   double max = -min_ > max_ ? -min_ : max_;
   glEnable(GL_LINE_STIPPLE);
   glLineStipple(1,0xaaaa);
   glColor3f(0.5,0.5,0.5);
   glBegin(GL_LINES);
      glVertex3f(-1.1,0,0);
      glVertex3f(1.1,0,0);
   glEnd();
   glBegin(GL_LINES);
      glVertex3f(0,-1.1,0);
      glVertex3f(0,1.1,0);
   glEnd();
   glBegin(GL_LINES);
      glVertex3f(0,0,-max - 0.1);
      glVertex3f(0,0,max + 0.1);
   glEnd();
   glDisable(GL_LINE_STIPPLE);
   glRasterPos3f(1.3,0,0);
   glBitmap(8, 13, 0.0, 2.0, 10.0, 0.0, x_bits);
   glRasterPos3f(0,1.3,0);
   glBitmap(8, 13, 0.0, 2.0, 10.0, 0.0, y_bits);
   glRasterPos3f(0,0,max + 0.3);
   glBitmap(8, 13, 0.0, 2.0, 10.0, 0.0, z_bits);
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::drawGrid()
{
   trace("Graph3d::drawGrid");
   printf("drawGrid()\n");
   glDrawElements(GL_QUADS, (GW-1) * (GH-1) * 4, GL_UNSIGNED_INT, indices_);
}

//
// Draw the complete stuff.
//
template <int GW,int GH>
void graph::Graph3d<GW,GH>::display()
{
   trace("Graph3d::display");
   if ( (mode_ & DLISTS) && recompile_ ) {
      createDisplayList();
      recompile_ = false;
   }
   glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glPushMatrix ();
      glTranslatef (0.0, 0.0, tz_);
      glMultMatrixf (modelview_);
      if ( mode_ & DLISTS ) {
         glCallList(hgraph_);
      }
      else {
         drawGrid();
      }
      if ( mode_ & AXES ) {
         if ( mode_ & LIGHTS ) {
            glDisable(GL_LIGHTING);
         }
         if ( mode_ & DLISTS ) {
            glCallList(haxes_);
         }
         else {
            drawAxes();
         }
         if ( mode_ & LIGHTS ) {
            glEnable(GL_LIGHTING);
         }
      }
   glPopMatrix ();
}

//
// This reacts to window size changes.
//
template <int GW,int GH>
void graph::Graph3d<GW,GH>::reshape(int w,int h)
{
   trace("Graph3d::reshape");
   glViewport (0, 0, (GLsizei) w, (GLsizei) h);
   glMatrixMode (GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(40.0,(GLfloat) w/(GLfloat) h, 1.0, 20.0);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
}

template <int GW,int GH>
graph::Graph3d<GW,GH>::~Graph3d()
{
   trace("Graph3d::~Graph3d");
   rfunc_->unref();
   gfunc_->unref();
   bfunc_->unref();
   function_->unref();
   if ( lists_created_ ) {
      glDeleteLists(hgraph_,1);
      glDeleteLists(haxes_,1);
   }
   delete[] light0_ambient_;
   delete[] light0_color_;
   delete[] light1_ambient_;
   delete[] light1_color_;  
}

template <int GW,int GH>
float graph::Graph3d<GW,GH>::getLightRedValue(int n) const
{
   switch ( n ) {
      case 0:
         return light0_color_[0];
      case 1:
         return light1_color_[0];
      default:
         throw illegal_argument_error();
   }
}

template <int GW,int GH>
float graph::Graph3d<GW,GH>::getLightGreenValue(int n) const
{
   switch ( n ) {
      case 0:
         return light0_color_[1];
      case 1:
         return light1_color_[1];
      default:
         throw illegal_argument_error();
   }
}

template <int GW,int GH>
float graph::Graph3d<GW,GH>::getLightBlueValue(int n) const
{
   switch ( n ) {
      case 0:
         return light0_color_[2];
      case 1:
         return light1_color_[2];
      default:
         throw illegal_argument_error();
   }
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::screenRotate(int x0,int y0,int x1,int y1)
{
   trace("Graph3d::screenRotate");
   GLint vp[4];
   GLdouble pmat[16], mvmat[16];
   GLdouble w0[3], w1[3], axis[3];
   glGetIntegerv(GL_VIEWPORT,vp);
   glGetDoublev(GL_PROJECTION_MATRIX,pmat);
   glGetDoublev(GL_MODELVIEW_MATRIX,mvmat);
   GLint realy0 = vp[3] - (GLint)y0 - 1,
         realy1 = vp[3] - (GLint)y1 - 1;
   gluUnProject( (GLdouble)x0, (GLdouble) realy0, 0.0,
                 mvmat, pmat, vp, w0, w0+1, w0+2 );
   gluUnProject( (GLdouble)x1, (GLdouble) realy1, 0.0,
                 mvmat, pmat, vp, w1, w1+1, w1+2 );
   crossprod(w0,w1,axis);
   GLdouble w0_abs   = sqrt(w0[0] * w0[0] +
                            w0[1] * w0[1] +
                            w0[2] * w0[2]);
   GLdouble w1_abs   = sqrt(w1[0] * w1[0] +
                            w1[1] * w1[1] +
                            w1[2] * w1[2]);
   GLdouble axis_abs = sqrt(axis[0] * axis[0] +
                            axis[1] * axis[1] +
                            axis[2] * axis[2]);
   GLdouble angle = acos(axis_abs / (w0_abs * w1_abs));
   glPushMatrix();
      glLoadMatrixf(modelview_);
      glRotatef(angle,axis[0],axis[1],axis[2]);
      glGetFloatv(GL_MODELVIEW_MATRIX,modelview_);
   glPopMatrix();
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::setClearColor(float r,float g,float b)
{
   bg_r_ = r;
   bg_g_ = g;
   bg_b_ = b;
   glClearColor(r,g,b,0);
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::setLightColor(int n,float r,float g,float b)
{
   switch ( n ) {
      case 0:
         light0_ambient_[0] = r / 5;
         light0_ambient_[1] = g / 5;
         light0_ambient_[2] = b / 5;
         light0_ambient_[3] = 1.0;
         light0_color_[0] = r;
         light0_color_[1] = g;
         light0_color_[2] = b;
         light0_color_[3] = 1.0;
         glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambient_);
         glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_color_);
         glLightfv(GL_LIGHT0, GL_SPECULAR, light0_color_);
	 break;
      case 1:
         light1_ambient_[0] = r / 5;
         light1_ambient_[1] = g / 5;
         light1_ambient_[2] = b / 5;
         light1_ambient_[3] = 1.0;
         light1_color_[0] = r;
         light1_color_[1] = g;
         light1_color_[2] = b;
         light1_color_[3] = 1.0;
         glLightfv(GL_LIGHT1, GL_AMBIENT, light1_ambient_);
         glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_color_);
         glLightfv(GL_LIGHT1, GL_SPECULAR, light1_color_);
	 break;
      default:
         throw illegal_argument_error();    
   }
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::setMode(int mode)
{
   mode_ |= mode;
   if ( mode_ & POINTS ) {
      setPoints();
   }
   if ( mode_ & LINES ) {
      setLines();
   }
   if ( mode_ & SMOOTH ) {
      setSmooth();
   }
   setLighting( mode_ & LIGHTS );
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::setDefaultMode()
{
   mode_ = POINTS;
   setPoints();
   setLighting(false);
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::setPoints()
{
   glPolygonMode(GL_FRONT_AND_BACK,GL_POINT);
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::setLines()
{
   glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::setSmooth()
{
   glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::rotateX(double rx)
{
   rx_ = rx;
   glPushMatrix();
      glLoadMatrixf(modelview_);
      glRotatef(rx_,1.0,0.0,0.0);
      glGetFloatv(GL_MODELVIEW_MATRIX,modelview_);
   glPopMatrix();
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::rotateY(double ry)
{
   ry_ = ry;
   glPushMatrix();
      glLoadMatrixf(modelview_);
      glRotatef(ry_,0.0,1.0,0.0);
      glGetFloatv(GL_MODELVIEW_MATRIX,modelview_);
   glPopMatrix();
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::rotateZ(double rz)
{
   rz_ = rz;
   glPushMatrix();
      glLoadMatrixf(modelview_);
      glRotatef(rz_,0.0,0.0,1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX,modelview_);
   glPopMatrix();
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::setLighting(bool lighting)
{
   if ( lighting ) {
      glEnable(GL_LIGHTING);
   }
   else {
      glDisable(GL_LIGHTING);
   }
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::setRFunction(graph::Function3& rfunc)
{
   if ( rfunc_ ) {
      rfunc_->unref();
   }
   rfunc_ = &rfunc;
   rfunc_->ref();
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::setGFunction(graph::Function3& gfunc)
{
   if ( gfunc_ ) {
      gfunc_->unref();
   }
   gfunc_ = &gfunc;
   gfunc_->ref();
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::setBFunction(graph::Function3& bfunc)
{
   if ( bfunc_ ) {
      bfunc_->unref();
   }
   bfunc_ = &bfunc;
   bfunc_->ref();
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::unsetMode(int mode)
{
   mode_ &= ~mode;
   if ( mode_ & POINTS ) {
      setPoints();
   }
   if ( mode_ & LINES ) {
      setLines();
   }
   if ( mode_ & SMOOTH ) {
      setSmooth();
   }
   setLighting( mode_ & LIGHTS );
}

template <int GW,int GH>
void graph::Graph3d<GW,GH>::zoom(double factor)
{
   tz_ += factor;
}

