#ifndef _jgraph_h_
#define _jgraph_h_

// (C) by Thalion Graphic Solutions e.K. 2001
//
// Permissions to copy, modify and redistribute
// granted, according to the conditions defined
// in the Gnu Public License (GPL).
// Please read the file gpl.txt provided with this
// distribution.

#include <graph/graph3d.h>
#include <graph/runtime.h>

#include <GL/glx.h>
#include <jni.h>

namespace graph {

   typedef Graph3d<100,100> graph3d_t;

   //
   // This class allows access of a graph3d.Function2
   // instance.
   //
   class JGraphable2: public GraphableFunction2<100,100> {
      public:
         friend class JGraphFactory;
	 double operator () (double,double);
	 Graph3d<100,100> *getGraph(double x0=-1,
                                    double x1=1,
                                    double y0=-1,
                                    double y1=1,
                                    int mode=graph3d_t::POINTS|graph3d_t::DLISTS);
         ~JGraphable2();
      private:
         // Accessible for factory only.
         JGraphable2(JNIEnv *,jobject);
	 // Not defined.
	 JGraphable2(const JGraphable2&);
	 JGraphable2(JGraphable2&);
	 JGraphable2& operator=(const JGraphable2&);
	 // Attributes.
	 JNIEnv       *env_;
         jobject       jfunc_;
         jmethodID     mid_;
   };
   
   //
   // This class allows access of a graph3d.Function3
   // instance.
   //
   class JGraphable3: public GraphableFunction3<100,100> {
      public:
         friend class JGraphFactory;
	 double operator () (double,double,double);
	 Graph3d<100,100> *getGraph(double x0=-1,
                                    double x1=1,
                                    double y0=-1,
                                    double y1=1,
                                    int mode=graph3d_t::POINTS|graph3d_t::DLISTS);
	 GraphableFunction3<100,100> *clone() const;	 
         ~JGraphable3();
      private:
         // Accessible for factory only.
         JGraphable3(JNIEnv *,jobject);
	 // Not defined.
	 JGraphable3(const JGraphable2&);
	 JGraphable3(JGraphable2&);
	 JGraphable3& operator=(const JGraphable2&);
	 // Attributes.
	 JNIEnv       *env_;	 
         jobject       jfunc_;
         jmethodID     mid_;      
   };
   
   //
   // The class extends Graph3d by JNI specific stuff.
   // It uses the grid default dimensions for the graph.
   //
   class JGraph3d: public Graph3d<100,100> {

      public:
      
         // Plotting modes
	 enum mode_bits {
	    CONTOURS = 0x40
	 };
      
	 // No immediate creation of a JGraph3d.
         friend class JGraphFactory;
	 friend class JGraphable2;
	 friend class JGraphable3;

         // Initialize color parameters different from
         // defaults.
         void initColors(float,float,float,
                         float,float,float,
                         float,float,float);

	 void realize(Display *,
                      XVisualInfo *,
                      GLXDrawable);
		      
         void display(GLXDrawable);

	 void reshape(GLXDrawable,
                      int,
                      int);

         void print(XVisualInfo *,
                    GLXDrawable,
                    int,
                    int);
	 
	 // Get light colors.
	 float getLightRedValue(int) const;
	 float getLightGreenValue(int) const;
	 float getLightBlueValue(int) const;

	 // Plot ranges.
	 double getRangeX0() const;
	 double getRangeX1() const;
	 double getRangeY0() const;
	 double getRangeY1() const;

         static graph::Runtime& getRuntime();
	 
	 // Is it a f(x,y,z)?
	 bool hasZ() const;
	 
	 // Get mode.
	 int mode(int) const;
	 
	 // Rotate around axis given by pick coordinates.
	 void screenRotate(GLXDrawable,int,int,int,int);


         // Set plot X and Y ranges.
         void setRangeX(GLXDrawable,double,double);
         void setRangeY(GLXDrawable,double,double);

         // Set Z value for a f(x,y,z) style function.
         // Will be ignored for binary function.
         void setZ(double);

         // Set background/clear color.
         void setClearColor(GLXDrawable,float,float,float);

         // Set light #n color.
         void setLightColor(GLXDrawable,int,float,float,float);

         // Set shading mode.
	 void setDefaultMode(GLXDrawable);
         void setMode(GLXDrawable,int);
	 void unsetMode(GLXDrawable,int);

         // Rotate around x,y,z axis.
         void rotateX(GLXDrawable,double);
         void rotateY(GLXDrawable,double);
         void rotateZ(GLXDrawable,double);
	 
	 // Zoom in & out.
	 void zoom(GLXDrawable,double);              
		      
         ~JGraph3d();
		 
      private:
      
         // Create the graph.
         JGraph3d(GraphableFunction2<100,100>& f,
                  double x0=-1,
                  double x1=1,
                  double y0=-1,
                  double y1=1,
                  int mode=POINTS|DLISTS);
         JGraph3d(GraphableFunction3<100,100>& f,
                  double x0=-1,
                  double x1=1,
                  double y0=-1,
                  double y1=1,
                  int mode=POINTS|DLISTS);
      
         // My display.
	 Display              *dpy_;
      
         // My GLX context.
         GLXContext            context_;

         // The stack machine runtime environment
         // to use for interpreted functions.
         static graph::Runtime rt_;

   };
   
   //
   // Factory class for graph instances.
   // This class takes care for creating the suitable
   // JGraphable* object and passing it to a JGraph3d.
   //
   class JGraphFactory {
      public:
      
	 class initialization_error {};
	 class typemismatch_error {};
	 
         static JGraph3d *create(JNIEnv *env,
	                         jobject jfunc,
				 double x0=-1,
				 double x1=1,
				 double y0=-1,
				 double y1=1,
                                 int mode=graph3d_t::POINTS|graph3d_t::DLISTS);
      private:
         static jclass    jfunc2_clazz_;
	 static jclass    jfunc3_clazz_;
	 static jclass    interpfunc_clazz_;
	 static jfieldID  hfunction_id_;
	 static bool      initialized_;
   };

}

#endif
