// (C) by Thalion Graphic Solutions e.K. 2001
//
// Permissions to copy, modify and redistribute
// granted, according to the conditions defined
// in the Gnu Public License (GPL).
// Please read the file gpl.txt provided with this
// distribution.

#include "graph3d_peer_Context.h"
#include "jgraph.h"

#include <graph/codegen.h>

#include <GL/glu.h>
#include <assert.h>
#include <jawt.h>
#include <jawt_md.h>
#include <stdlib.h>

#include "trace.h"

using graph::CodeGen;
using graph::Function;
using graph::JGraphFactory;
using graph::JGraph3d;
using graph::Runtime;

//////////////////////////////////////////////////////////////////
//
// Global utility stuff.
//

//
// The class manages canvas access and locking via
// JAWT. Note: in this application, the canvas lock also
// serves as a global lock for the non-reentrant GL application.
// This works with X11, but probably not with Win32, or another
// UI.
// We needed an extra library lock in this case.
//
class JawtMgr {
   public:
      JawtMgr(JNIEnv *,jobject);
      bool good() const;
      JAWT_X11DrawingSurfaceInfo *getDsi();
      ~JawtMgr();
   private:
      JAWT                     awt_; 
      JAWT_DrawingSurface     *ds_;
      JAWT_DrawingSurfaceInfo *dsi_;
      bool                  good_;
};

//
// The selected viewport and offscreen visuals.
//
static XVisualInfo *vp_vis = NULL;
static XVisualInfo *offs_vis = NULL;

//
// These are handles to the Java World attributes which must be
// accessible from here. They will be pre-initialized in the
// initHandle() method.
//
jmethodID hgetWidth, hgetHeight;

//
// Method handle for sun.awt.X11GraphicsConfig.getVisual().
//
static jmethodID hGetVisual;

//
// The code generator.
//
static CodeGen *codegen = NULL;

//
// Java exception classes used.
//
static jclass databuffer_exceeded_class;
static jclass illegal_argument_exception_class;
static jclass nullpointer_exception_class;
static jclass runtime_exception_class;
static jclass typemismatch_exception_class;

//
// JAWT manager method definitions.
//
inline JawtMgr::JawtMgr(JNIEnv *env,jobject canvas):
   ds_(NULL),dsi_(NULL),good_(false)
{
   trace("JawtMgr::JawtMgr");
   jboolean result;
   jint lock;
   awt_.version = JAWT_VERSION_1_3;
   result = JAWT_GetAWT(env, &awt_);
   assert(result != JNI_FALSE);
   ds_ = awt_.GetDrawingSurface(env, canvas);
   assert(ds_ != NULL);
   lock = ds_->Lock(ds_);
   good_ =  ! (lock & JAWT_LOCK_ERROR);
   if ( good_ ) {
      dsi_ = ds_->GetDrawingSurfaceInfo(ds_);
   }
   else {
      env->ThrowNew(illegal_argument_exception_class,"Native canvas lock failed");
   }
}

inline JAWT_X11DrawingSurfaceInfo *JawtMgr::getDsi()
{
   trace("JawtMgr::getDsi");
   return (JAWT_X11DrawingSurfaceInfo*)dsi_->platformInfo;
}

inline bool JawtMgr::good() const
{
   return good_;
}

inline JawtMgr::~JawtMgr()
{
   trace("JawtMgr::~JawtMgr");
   if ( good_ ) {
      ds_->FreeDrawingSurfaceInfo(dsi_);
      ds_->Unlock(ds_);
   }
   awt_.FreeDrawingSurface(ds_);
}

//////////////////////////////////////////////////////////////////
//
// JNI method definitions.
//

/*
 * Class:     graph3d_peer_Context
 * Method:    initLibrary
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_initLibrary
  (JNIEnv *env, jclass clazz)
{
   trace("Java_graph3d_peer_Context_initLibrary");
   jclass canvas_clazz = env->FindClass("java/awt/Canvas");
   assert( canvas_clazz != 0);
   hgetWidth = env->GetMethodID(canvas_clazz,"getWidth","()I");
   hgetHeight = env->GetMethodID(canvas_clazz,"getWidth","()I");
   codegen = new CodeGen(JGraph3d::getRuntime());
   databuffer_exceeded_class = env->FindClass("graph3d/peer/DataBufferExceeded");
   assert( databuffer_exceeded_class != NULL );
   databuffer_exceeded_class = (jclass) env->NewGlobalRef(databuffer_exceeded_class);
   illegal_argument_exception_class = env->FindClass("java/lang/IllegalArgumentException");
   assert( illegal_argument_exception_class != NULL );
   illegal_argument_exception_class = (jclass) env->NewGlobalRef(illegal_argument_exception_class);
   nullpointer_exception_class = env->FindClass("java/lang/NullPointerException");
   assert( nullpointer_exception_class != NULL );
   nullpointer_exception_class = (jclass) env->NewGlobalRef(nullpointer_exception_class);
   runtime_exception_class = env->FindClass("graph3d/RuntimeException");
   assert( runtime_exception_class != NULL );
   runtime_exception_class = (jclass) env->NewGlobalRef(runtime_exception_class);
   typemismatch_exception_class = env->FindClass("graph3d/TypeMismatchException");
   assert( typemismatch_exception_class != NULL );
   typemismatch_exception_class = (jclass) env->NewGlobalRef(typemismatch_exception_class);
}

/*
 * Class:     graph3d_peer_Context
 * Method:    findConfiguration
 * Signature: ()Z
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_findConfiguration
  (JNIEnv *env, jobject self)
{
   trace("Java_graph3d_peer_Context_findConfiguration");
   //
   // Initialize the method handle.
   //
   jclass gc_clazz = env->FindClass("sun/awt/X11GraphicsConfig");
   assert( gc_clazz != 0 );
   hGetVisual = env->GetMethodID(gc_clazz,"getVisual","()I");
   //
   // Find the visuals we want to use.
   //
   char *dpyname = getenv("DISPLAY");
   Display *dpy = XOpenDisplay(dpyname ? dpyname : "localhost:0.0");
   int vp_attribList[] = { GLX_RGBA,
                           GLX_DOUBLEBUFFER,
                           GLX_RED_SIZE, 3,
	 		   GLX_DEPTH_SIZE, 4,
                           None };
   vp_vis =
      glXChooseVisual(dpy, XScreenNumberOfScreen(DefaultScreenOfDisplay(dpy)), vp_attribList);
   if ( ! vp_vis ) {
      env->ThrowNew(nullpointer_exception_class, "Viewport visual");
   }

   int offs_attribList[] = { GLX_RGBA,
                             GLX_RED_SIZE, 8, 
                             GLX_GREEN_SIZE, 8,
                             GLX_BLUE_SIZE, 8,
	 		     GLX_DEPTH_SIZE, 1,
                             None};
   offs_vis =
      glXChooseVisual(dpy, XScreenNumberOfScreen(DefaultScreenOfDisplay(dpy)), offs_attribList);
   if ( ! vp_vis ) {
      env->ThrowNew(nullpointer_exception_class, "Offscreen visual");
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    checkOffsConfiguration
 * Signature: (Ljava/awt/GraphicsConfiguration;)Z
 */
JNIEXPORT jboolean JNICALL Java_graph3d_peer_Context_checkOffsConfiguration
  (JNIEnv *env, jobject self, jobject gc)
{
   trace("Java_graph3d_peer_Context_checkOffsConfiguration");
   return env->CallIntMethod(gc,hGetVisual) == offs_vis->visualid ? JNI_TRUE : JNI_FALSE;
}

/*
 * Class:     graph3d_peer_Context
 * Method:    checkVpConfiguration
 * Signature: (Ljava/awt/GraphicsConfiguration;)Z
 */
JNIEXPORT jboolean JNICALL Java_graph3d_peer_Context_checkVpConfiguration
  (JNIEnv *env, jobject self, jobject gc)
{
   trace("Java_graph3d_peer_Context_checkVpConfiguration");
   return env->CallIntMethod(gc,hGetVisual) == vp_vis->visualid ? JNI_TRUE : JNI_FALSE;
}

/*
 * Class:     graph3d_peer_Context
 * Method:    create
 * Signature: (Lgraph3d/Function;DDDDI)J
 */
JNIEXPORT jlong JNICALL Java_graph3d_peer_Context_create
  (JNIEnv *env, jobject self, jobject jfunction, jdouble x0, jdouble x1, jdouble y0, jdouble y1, jint mode)
{
   trace("Java_graph3d_peer_Context_create");
   try {
      return (jlong) JGraphFactory::create(env,jfunction,x0,x1,y0,y1,mode);
   }
   catch ( JGraphFactory::initialization_error ) {
      env->ThrowNew(nullpointer_exception_class, "initialization");
   }
   catch ( JGraph3d::illegal_argument_error ) {
      env->ThrowNew(illegal_argument_exception_class, "x0>=x1 || y0>=y1");
   }
   catch ( JGraphFactory::typemismatch_error ) {
      env->ThrowNew(typemismatch_exception_class, "function");
   }
   catch ( Runtime::illegal_instruction ) {
      env->ThrowNew(runtime_exception_class, "illegal instruction");     
   }
   catch ( Runtime::division_by_zero ) {
      env->ThrowNew(runtime_exception_class, "division by zero");        
   }
   catch ( Runtime::overflow ) {
      env->ThrowNew(runtime_exception_class, "stack overflow");        
   }
   catch ( Runtime::underflow ) {
      env->ThrowNew(runtime_exception_class, "stack underflow");        
   }
   printf("*** Exception raised from JGraphFactory::create!\n");
   return (jlong) 0;
}

/*
 * Class:     graph3d_peer_Context
 * Method:    initColors
 * Signature: (JFFFFFFFFF)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_initColors
  (JNIEnv *env, jobject self, jlong hgraph,
   jfloat l0r, jfloat l0g, jfloat l0b, jfloat l1r, jfloat l1g, jfloat l1b, jfloat bgr, jfloat bgg, jfloat bgb)
{
   trace("Java_graph3d_peer_Context_initColors");
   assert( hgraph != 0 );
   ((JGraph3d *)hgraph)->initColors(l0r,l0g,l0b,l1r,l1g,l1b,bgr,bgg,bgb);
}

/*
 * Class:     graph3d_peer_Context
 * Method:    realize
 * Signature: (JLjava/awt/Canvas;)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_realize
  (JNIEnv *env, jobject self, jlong hgraph, jobject canvas)
{
   trace("Java_graph3d_peer_Context_realize");
   assert( hgraph != 0 );
   JawtMgr mgr(env,canvas);
   if ( ! mgr.good() ) {
      return;
   }
   if ( vp_vis ) {
      ((JGraph3d *)hgraph)->realize(mgr.getDsi()->display,vp_vis,mgr.getDsi()->drawable);
      ((JGraph3d *)hgraph)->reshape(mgr.getDsi()->drawable,
                                    env->CallIntMethod(canvas,hgetWidth),
                                    env->CallIntMethod(canvas,hgetHeight));
   }
   else {
      env->ThrowNew(nullpointer_exception_class, "Viewport visual");
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    display
 * Signature: (JLjava/awt/Canvas;)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_display
  (JNIEnv *env, jobject self, jlong hgraph, jobject canvas)
{
   trace("Java_graph3d_peer_Context_display");
   JawtMgr mgr(env,canvas);
   if ( mgr.good() ) {
      ((JGraph3d *)hgraph)->display(mgr.getDsi()->drawable);
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    reshape
 * Signature: (JLjava/awt/Canvas;II)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_reshape
  (JNIEnv *env, jobject self, jlong hgraph, jobject canvas, jint width, jint height)
{
   trace("Java_graph3d_peer_Context_reshape");
   JawtMgr mgr(env,canvas);
   if ( mgr.good() ) {
      ((JGraph3d *)hgraph)->reshape(mgr.getDsi()->drawable,width,height);
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    getImageData
 * Signature: (JLjava/awt/Canvas;II)[I
 */
JNIEXPORT jintArray JNICALL Java_graph3d_peer_Context_getImageData
  (JNIEnv *env,
   jobject self,
   jlong hgraph,
   jobject canvas,
   jint w,
   jint h)
{
   trace("Java_graph3d_peer_Context_getImageData");
   JawtMgr mgr(env,canvas);
   jintArray int_array = NULL;
   if ( mgr.good() ) {
      if ( offs_vis ) {
         Pixmap pixmap =
            XCreatePixmap(mgr.getDsi()->display,mgr.getDsi()->drawable,w,h,offs_vis->depth);
         GLXPixmap glx_pixmap = glXCreateGLXPixmap(mgr.getDsi()->display,offs_vis,pixmap);

         printf("visual depth = %d\n",offs_vis->depth);

         ((JGraph3d *)hgraph)->print(offs_vis,glx_pixmap,w,h);
         XSync(mgr.getDsi()->display,false);

         XImage *image = XGetImage(mgr.getDsi()->display,pixmap,0,0,w,h,AllPlanes,ZPixmap);
         XSync(mgr.getDsi()->display,false);

         size_t size = w * h;
         int_array = env->NewIntArray(size);
         env->SetIntArrayRegion(int_array,0,size,(jint *)image->data);
         XDestroyImage(image);

         glXDestroyGLXPixmap(mgr.getDsi()->display,glx_pixmap);
         XFreePixmap(mgr.getDsi()->display,pixmap);
      }
      else {
         env->ThrowNew(nullpointer_exception_class, "Offscreen visual");
      }
   }
   return int_array;
}

/*
 * Class:     graph3d_peer_Context
 * Method:    getLightRedValue
 * Signature: (JI)F
 */
JNIEXPORT jfloat JNICALL Java_graph3d_peer_Context_getLightRedValue
  (JNIEnv *env, jobject self, jlong hgraph, jint n)
{
   trace("Java_graph3d_peer_Context_getLightRedValue");
   try {
      return (jfloat) ((JGraph3d *)hgraph)->getLightRedValue(n);
   }
   catch ( JGraph3d::illegal_argument_error ) {
      env->ThrowNew(illegal_argument_exception_class, "n");
      return (jfloat) 0;   
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    getLightGreenValue
 * Signature: (JI)F
 */
JNIEXPORT jfloat JNICALL Java_graph3d_peer_Context_getLightGreenValue
  (JNIEnv *env, jobject self, jlong hgraph, jint n)
{
   trace("Java_graph3d_peer_Context_getLightGreenValue");
   try {
      return (jfloat) ((JGraph3d *)hgraph)->getLightGreenValue(n);
   }
   catch ( JGraph3d::illegal_argument_error ) {
      env->ThrowNew(illegal_argument_exception_class, "n");
      return (jfloat) 0;   
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    getLightBlueValue
 * Signature: (JI)F
 */
JNIEXPORT jfloat JNICALL Java_graph3d_peer_Context_getLightBlueValue
  (JNIEnv *env, jobject self, jlong hgraph, jint n)
{
   trace("Java_graph3d_peer_Context_getLightBlueValue");
   try {
      return (jfloat) ((JGraph3d *)hgraph)->getLightBlueValue(n);
   }
   catch ( JGraph3d::illegal_argument_error ) {
      env->ThrowNew(illegal_argument_exception_class, "n");
      return (jfloat) 0;   
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    getRangeX0
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_graph3d_peer_Context_getRangeX0
  (JNIEnv *env, jobject self, jlong hgraph)
{
   trace("Java_graph3d_peer_Context_getRangeX0");
   return (jdouble) ((JGraph3d *)hgraph)->getRangeX0();
}

/*
 * Class:     graph3d_peer_Context
 * Method:    getRangeX1
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_graph3d_peer_Context_getRangeX1
  (JNIEnv *env, jobject self, jlong hgraph)
{
   trace("Java_graph3d_peer_Context_getRangeX1");
   return (jdouble) ((JGraph3d *)hgraph)->getRangeX1();
}

/*
 * Class:     graph3d_peer_Context
 * Method:    getRangeY0
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_graph3d_peer_Context_getRangeY0
  (JNIEnv *env, jobject self, jlong hgraph)
{
   trace("Java_graph3d_peer_Context_getRangeY0");
   return (jdouble) ((JGraph3d *)hgraph)->getRangeY0();
}

/*
 * Class:     graph3d_peer_Context
 * Method:    getRangeY1
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_graph3d_peer_Context_getRangeY1
  (JNIEnv *env, jobject self, jlong hgraph)
{
   trace("Java_graph3d_peer_Context_getRangeY1");
   return (jdouble) ((JGraph3d *)hgraph)->getRangeY1();
}

/*
 * Class:     graph3d_peer_Context
 * Method:    hasZ
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_graph3d_peer_Context_hasZ
  (JNIEnv *env, jobject self, jlong hgraph)
{
   trace("Java_graph3d_peer_Context_hasZ");
   return (jboolean) ((JGraph3d *)hgraph)->hasZ();
}

/*
 * Class:     graph3d_peer_Context
 * Method:    mode
 * Signature: (JI)I
 */
JNIEXPORT jint JNICALL Java_graph3d_peer_Context_mode
  (JNIEnv *env, jobject self, jlong hgraph, jint m)
{
   trace("Java_graph3d_peer_Context_mode");
   return (jint) ((JGraph3d *)hgraph)->mode(m);
}


/*
 * Class:     graph3d_peer_Context
 * Method:    screenRotate
 * Signature: (JLjava/awt/Canvas;IIII)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_screenRotate
  (JNIEnv *env, jobject self, jlong hgraph, jobject canvas, jint x0, jint y0, jint x1, jint y1)
{
   trace("Java_graph3d_peer_Context_screenRotate");
   JawtMgr mgr(env,canvas);
   if ( mgr.good() ) {
      ((JGraph3d *)hgraph)->screenRotate(mgr.getDsi()->drawable,x0,y0,x1,y1);
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    setRangeX
 * Signature: (JLjava/awt/Canvas;DD)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_setRangeX
  (JNIEnv *env, jobject self, jlong hgraph, jobject canvas, jdouble x0, jdouble x1)
{
   trace("Java_graph3d_peer_Context_setRangeX");
   JawtMgr mgr(env,canvas);
   if ( ! mgr.good() ) {
      return;
   }
   try {
      ((JGraph3d *)hgraph)->setRangeX(mgr.getDsi()->drawable,x0,x1);
   }
   catch ( JGraph3d::illegal_argument_error ) {
      env->ThrowNew(illegal_argument_exception_class, "x0>=x1");  
   }
   catch ( Runtime::illegal_instruction ) {
      env->ThrowNew(runtime_exception_class, "illegal instruction");     
   }
   catch ( Runtime::division_by_zero ) {
      env->ThrowNew(runtime_exception_class, "division by zero");        
   }
   catch ( Runtime::overflow ) {
      env->ThrowNew(runtime_exception_class, "stack overflow");        
   }
   catch ( Runtime::underflow ) {
      env->ThrowNew(runtime_exception_class, "stack underflow");        
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    setRangeY
 * Signature: (JLjava/awt/Canvas;DD)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_setRangeY
  (JNIEnv *env, jobject self, jlong hgraph, jobject canvas, jdouble y0, jdouble y1)
{
   trace("Java_graph3d_peer_Context_setRangeY");
   JawtMgr mgr(env,canvas);
   if ( ! mgr.good() ) {
      return;
   }
   try {
      ((JGraph3d *)hgraph)->setRangeY(mgr.getDsi()->drawable,y0,y1);
   }
   catch ( JGraph3d::illegal_argument_error ) {
      env->ThrowNew(illegal_argument_exception_class, "y0>=y1");  
   }
   catch ( Runtime::illegal_instruction ) {
      env->ThrowNew(runtime_exception_class, "illegal instruction");     
   }
   catch ( Runtime::division_by_zero ) {
      env->ThrowNew(runtime_exception_class, "division by zero");        
   }
   catch ( Runtime::overflow ) {
      env->ThrowNew(runtime_exception_class, "stack overflow");        
   }
   catch ( Runtime::underflow ) {
      env->ThrowNew(runtime_exception_class, "stack underflow");        
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    setZ
 * Signature: (JD)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_setZ
  (JNIEnv *env, jobject self, jlong hgraph, jdouble z)
{
   trace("Java_graph3d_peer_Context_setZ");
   try {
      ((JGraph3d *)hgraph)->setZ(z);
   }
   catch ( Runtime::illegal_instruction ) {
      env->ThrowNew(runtime_exception_class, "illegal instruction");     
   }
   catch ( Runtime::division_by_zero ) {
      env->ThrowNew(runtime_exception_class, "division by zero");        
   }
   catch ( Runtime::overflow ) {
      env->ThrowNew(runtime_exception_class, "stack overflow");        
   }
   catch ( Runtime::underflow ) {
      env->ThrowNew(runtime_exception_class, "stack underflow");        
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    setBackgroundColor
 * Signature: (JLjava/awt/Canvas;FFF)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_setBackgroundColor
  (JNIEnv *env, jobject self, jlong hgraph, jobject canvas, jfloat r, jfloat g, jfloat b)
{
   trace("Java_graph3d_peer_Context_setBackgroundColor");
   JawtMgr mgr(env,canvas);
   if ( ! mgr.good() ) {
      return;
   }
   ((JGraph3d *)hgraph)->setClearColor(mgr.getDsi()->drawable,r,g,b);
}

/*
 * Class:     graph3d_peer_Context
 * Method:    setLightColor
 * Signature: (JLjava/awt/Canvas;IFFF)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_setLightColor
  (JNIEnv *env, jobject self, jlong hgraph, jobject canvas, jint n, jfloat r, jfloat g, jfloat b)
{
   trace("Java_graph3d_peer_Context_setLightColor");
   JawtMgr mgr(env,canvas);
   if ( ! mgr.good() ) {
      return;
   }
   try {
      ((JGraph3d *)hgraph)->setLightColor(mgr.getDsi()->drawable,n,r,g,b);
   }
   catch ( JGraph3d::illegal_argument_error ) {
      env->ThrowNew(illegal_argument_exception_class, "n");  
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    setDefaultMode
 * Signature: (JLjava/awt/Canvas;)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_setDefaultMode
  (JNIEnv *env, jobject self, jlong hgraph, jobject canvas)
{
   trace("Java_graph3d_peer_Context_setDefaultMode");
   JawtMgr mgr(env,canvas);
   if ( mgr.good() ) {
      ((JGraph3d *)hgraph)->setDefaultMode(mgr.getDsi()->drawable);
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    setMode
 * Signature: (JLjava/awt/Canvas;I)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_setMode
  (JNIEnv *env, jobject self, jlong hgraph, jobject canvas, jint mode)
{
   trace("Java_graph3d_peer_Context_setMode");
   JawtMgr mgr(env,canvas);
   if ( mgr.good() ) {
      ((JGraph3d *)hgraph)->setMode(mgr.getDsi()->drawable,mode);
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    rotateX
 * Signature: (JLjava/awt/Canvas;D)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_rotateX
  (JNIEnv *env, jobject self, jlong hgraph, jobject canvas, jdouble ax)
{
   trace("Java_graph3d_peer_Context_rotateX");
   JawtMgr mgr(env,canvas);
   if ( mgr.good() ) {
      ((JGraph3d *)hgraph)->rotateX(mgr.getDsi()->drawable,ax);
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    rotateY
 * Signature: (JLjava/awt/Canvas;D)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_rotateY
  (JNIEnv *env, jobject self, jlong hgraph, jobject canvas, jdouble ay)
{
   trace("Java_graph3d_peer_Context_rotateY");
   JawtMgr mgr(env,canvas);
   if ( mgr.good() ) {
      ((JGraph3d *)hgraph)->rotateY(mgr.getDsi()->drawable,ay);
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    rotateZ
 * Signature: (JLjava/awt/Canvas;D)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_rotateZ
  (JNIEnv *env, jobject self, jlong hgraph, jobject canvas, jdouble az)
{
   trace("Java_graph3d_peer_Context_rotateZ");
   JawtMgr mgr(env,canvas);
   if ( mgr.good() ) {
      ((JGraph3d *)hgraph)->rotateZ(mgr.getDsi()->drawable,az);
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    unsetMode
 * Signature: (JLjava/awt/Canvas;I)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_unsetMode
  (JNIEnv *env, jobject self, jlong hgraph, jobject canvas, jint mode)
{
   trace("Java_graph3d_peer_Context_unsetMode");
   JawtMgr mgr(env,canvas);
   if ( mgr.good() ) {
      ((JGraph3d *)hgraph)->unsetMode(mgr.getDsi()->drawable,mode);
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    zoom
 * Signature: (JLjava/awt/Canvas;D)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_zoom
  (JNIEnv *env, jobject self, jlong hgraph, jobject canvas, jdouble tz)
{
   trace("Java_graph3d_peer_Context_zoom");
   JawtMgr mgr(env,canvas);
   if ( mgr.good() ) {
      ((JGraph3d *)hgraph)->zoom(mgr.getDsi()->drawable,tz);
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    dispose
 * Signature: (JLjava/awt/Canvas;)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_dispose
  (JNIEnv *env, jobject self, jlong hgraph, jobject canvas)
{
   trace("Java_graph3d_peer_Context_dispose");
   delete (JGraph3d *)hgraph;
   XFree(vp_vis);
   XFree(offs_vis);
}

/*
 * Class:     graph3d_peer_Context
 * Method:    beginFunction
 * Signature: (JLjava/lang/String;Ljava/lang/String;I)I
 */
JNIEXPORT jint JNICALL Java_graph3d_peer_Context_beginFunction
  (JNIEnv *env, jobject self, jstring jname, jstring jsig, jint argc)
{
   trace("Java_graph3d_peer_Context_beginFunction");
   const char *name = env->GetStringUTFChars(jname, 0); 
   const char *sig = env->GetStringUTFChars(jsig, 0);  
   jint id = (jint) codegen->beginFunction(name,sig,argc);
   env->ReleaseStringUTFChars(jsig,sig);
   env->ReleaseStringUTFChars(jname,name);
   return id;
}

/*
 * Class:     graph3d_peer_Context
 * Method:    addIns
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_graph3d_peer_Context_addIns__I
  (JNIEnv *env, jobject self, jint ins)
{
   trace("Java_graph3d_peer_Context_addIns");
   return (jint) codegen->addIns((Runtime::ins_t)ins);
}

/*
 * Class:     graph3d_peer_Context
 * Method:    addIns
 * Signature: (II)I
 */
JNIEXPORT jint JNICALL Java_graph3d_peer_Context_addIns__II
  (JNIEnv *env, jobject self, jint ins, jint arg)
{
   trace("Java_graph3d_peer_Context_addIns");
   return (jint) codegen->addIns((Runtime::ins_t)ins,(unsigned)arg);
}

/*
 * Class:     graph3d_peer_Context
 * Method:    setLabel
 * Signature: (II)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_setLabel
  (JNIEnv *env, jobject self, jint addr, jint label)
{
   trace("Java_graph3d_peer_Context_addIns");
   codegen->setLabel((unsigned)addr,(unsigned)label);
}

/*
 * Class:     graph3d_peer_Context
 * Method:    endFunction
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_endFunction
  (JNIEnv *env, jobject self)
{
   trace("Java_graph3d_peer_Context_endFunction");
   try {
      codegen->endFunction();
   }
   catch ( CodeGen::datasize_exceeded ) {
      env->ThrowNew(databuffer_exceeded_class,codegen->getName());    
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    reset
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_reset
  (JNIEnv *env, jobject self)
{
   trace("Java_graph3d_peer_Context_reset");
   codegen->reset();
}

/*
 * Class:     graph3d_peer_Context
 * Method:    getFunction
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_graph3d_peer_Context_getFunction
  (JNIEnv *env, jobject self)
{
   trace("Java_graph3d_peer_Context_getFunction");
   const Function *f = codegen->getFunction();
   f->print(stdout);
   return (jlong) f;
}


/*
 * Class:     graph3d_peer_Context
 * Method:    constant
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_graph3d_peer_Context_constant__J
  (JNIEnv *env, jobject self, jlong c)
{
   trace("Java_graph3d_peer_Context_constant");
   try {
      return (jint) codegen->constant((long)c);
   }
   catch ( CodeGen::datasize_exceeded ) {
      env->ThrowNew(databuffer_exceeded_class,codegen->getName());    
      return (jint) 0;   
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    constant
 * Signature: (D)I
 */
JNIEXPORT jint JNICALL Java_graph3d_peer_Context_constant__D
  (JNIEnv *env, jobject self, jdouble c)
{
   trace("Java_graph3d_peer_Context_constant");
   try {
      return (jint) codegen->constant((double)c);
   }
   catch ( CodeGen::datasize_exceeded ) {
      env->ThrowNew(databuffer_exceeded_class,codegen->getName());    
      return (jint) 0;   
   }
}

/*
 * Class:     graph3d_peer_Context
 * Method:    getFunctionName
 * Signature: (J)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_graph3d_peer_Context_getFunctionName
  (JNIEnv *env, jobject self, jlong hfunction)
{
   trace("Java_graph3d_peer_Context_getFunctionName");
   const char *name = ((Function *)hfunction)->getName();
   return env->NewStringUTF(name);
}

/*
 * Class:     graph3d_peer_Context
 * Method:    getFunctionType
 * Signature: (J)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_graph3d_peer_Context_getFunctionType
  (JNIEnv *env, jobject self, jlong hfunction)
{
   trace("Java_graph3d_peer_Context_getFunctionType");
   const char *type = ((Function *)hfunction)->getType();
   return env->NewStringUTF(type);
}

/*
 * Class:     graph3d_peer_Context
 * Method:    disposeFunction
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_graph3d_peer_Context_disposeFunction
  (JNIEnv *env, jobject self, jlong hfunction)
{
   trace("Java_graph3d_peer_Context_disposeFunction");
   delete (Function *)hfunction;
}


